from __future__ import annotations

import datetime as dt
from collections.abc import Callable
from functools import singledispatch
from typing import Any, cast

import hightime as ht

import nitypes.bintime as bt
from nitypes._exceptions import invalid_arg_type, invalid_requested_type
from nitypes.time.typing import AnyDateTime, AnyTimeDelta, TDateTime, TTimeDelta


def convert_datetime(requested_type: type[TDateTime], value: AnyDateTime, /) -> TDateTime:
    """Convert a datetime object to the specified type."""
    convert_func = _CONVERT_DATETIME_FOR_TYPE.get(requested_type)
    if convert_func is None:
        raise invalid_requested_type("datetime", requested_type)
    return cast(TDateTime, convert_func(value))


@singledispatch
def _convert_to_bt_absolute_time(value: object, /) -> bt.DateTime:
    raise invalid_arg_type("value", "datetime", value)


@_convert_to_bt_absolute_time.register
def _(value: bt.DateTime, /) -> bt.DateTime:
    return value


@_convert_to_bt_absolute_time.register
def _(value: dt.datetime, /) -> bt.DateTime:
    return bt.DateTime(value)


@_convert_to_bt_absolute_time.register
def _(value: ht.datetime, /) -> bt.DateTime:
    return bt.DateTime(value)


@singledispatch
def _convert_to_dt_datetime(value: object, /) -> dt.datetime:
    raise invalid_arg_type("value", "datetime", value)


@_convert_to_dt_datetime.register
def _(value: bt.DateTime, /) -> dt.datetime:
    return value._to_datetime_datetime()


@_convert_to_dt_datetime.register
def _(value: dt.datetime, /) -> dt.datetime:
    return value


@_convert_to_dt_datetime.register
def _(value: ht.datetime, /) -> dt.datetime:
    return dt.datetime(
        value.year,
        value.month,
        value.day,
        value.hour,
        value.minute,
        value.second,
        value.microsecond,
        value.tzinfo,
        fold=value.fold,
    )


@singledispatch
def _convert_to_ht_datetime(value: object, /) -> ht.datetime:
    raise invalid_arg_type("value", "datetime", value)


@_convert_to_ht_datetime.register
def _(value: bt.DateTime, /) -> ht.datetime:
    return value._to_hightime_datetime()


@_convert_to_ht_datetime.register
def _(value: dt.datetime, /) -> ht.datetime:
    return ht.datetime(
        value.year,
        value.month,
        value.day,
        value.hour,
        value.minute,
        value.second,
        value.microsecond,
        value.tzinfo,
        fold=value.fold,
    )


@_convert_to_ht_datetime.register
def _(value: ht.datetime, /) -> ht.datetime:
    return value


_CONVERT_DATETIME_FOR_TYPE: dict[type[Any], Callable[[object], object]] = {
    bt.DateTime: _convert_to_bt_absolute_time,
    dt.datetime: _convert_to_dt_datetime,
    ht.datetime: _convert_to_ht_datetime,
}


def convert_timedelta(requested_type: type[TTimeDelta], value: AnyTimeDelta, /) -> TTimeDelta:
    """Convert a timedelta object to the specified type."""
    convert_func = _CONVERT_TIMEDELTA_FOR_TYPE.get(requested_type)
    if convert_func is None:
        raise invalid_requested_type("timedelta", requested_type)
    return cast(TTimeDelta, convert_func(value))


@singledispatch
def _convert_to_bt_timedelta(value: object, /) -> bt.TimeDelta:
    raise invalid_arg_type("value", "timedelta", value)


@_convert_to_bt_timedelta.register
def _(value: bt.TimeDelta, /) -> bt.TimeDelta:
    return value


@_convert_to_bt_timedelta.register
def _(value: dt.timedelta, /) -> bt.TimeDelta:
    return bt.TimeDelta(value)


@_convert_to_bt_timedelta.register
def _(value: ht.timedelta, /) -> bt.TimeDelta:
    return bt.TimeDelta(value)


@singledispatch
def _convert_to_dt_timedelta(value: object, /) -> dt.timedelta:
    raise invalid_arg_type("value", "timedelta", value)


@_convert_to_dt_timedelta.register
def _(value: bt.TimeDelta, /) -> dt.timedelta:
    return value._to_datetime_timedelta()


@_convert_to_dt_timedelta.register
def _(value: dt.timedelta, /) -> dt.timedelta:
    return value


@_convert_to_dt_timedelta.register
def _(value: ht.timedelta, /) -> dt.timedelta:
    return dt.timedelta(value.days, value.seconds, value.microseconds)


@singledispatch
def _convert_to_ht_timedelta(value: object, /) -> ht.timedelta:
    raise invalid_arg_type("value", "timedelta", value)


@_convert_to_ht_timedelta.register
def _(value: bt.TimeDelta, /) -> ht.timedelta:
    return value._to_hightime_timedelta()


@_convert_to_ht_timedelta.register
def _(value: dt.timedelta, /) -> ht.timedelta:
    return ht.timedelta(
        value.days,
        value.seconds,
        value.microseconds,
    )


@_convert_to_ht_timedelta.register
def _(value: ht.timedelta, /) -> ht.timedelta:
    return value


_CONVERT_TIMEDELTA_FOR_TYPE: dict[type[Any], Callable[[object], object]] = {
    bt.TimeDelta: _convert_to_bt_timedelta,
    dt.timedelta: _convert_to_dt_timedelta,
    ht.timedelta: _convert_to_ht_timedelta,
}



"""Complex number data types for NI Python APIs.

================
Complex Integers
================

Some NI driver APIs (such as NI-FGEN, NI-SCOPE, NI-RFSA, and NI-RFSG) use complex numbers to
represent I/Q data. Python and NumPy have native support for complex floating-point numbers, but
not complex integers, so the :mod:`nitypes.complex` submodule provides a NumPy representation of
complex integers.

:any:`ComplexInt32DType` is a NumPy structured data type object representing a complex integer with
16-bit ``real`` and ``imag`` fields. This structured data type has the same memory layout as the
``NIComplexI16`` C struct used by NI driver APIs.

For more information about NumPy structured data types, see the
:ref:`NumPy documentation on structured arrays <numpy:structured_arrays>`.

.. note::
   In ``NIComplexI16``, the number 16 refers to the number of bits in each field. In
   :any:`ComplexInt32DType`, the number 32 refers to the total number of bits, following the
   precedent set by NumPy's other complex types. For example, :any:`numpy.complex128` contains
   64-bit ``real`` and ``imag`` fields.

Constructing arrays of complex integers
---------------------------------------

You can construct an array of complex integers from a sequence of tuples using :func:`numpy.array`:

>>> import numpy as np
>>> np.array([(1, 2), (3, 4)], dtype=ComplexInt32DType)
array([(1, 2), (3, 4)], dtype=[('real', '<i2'), ('imag', '<i2')])

Likewise, you can construct an array of complex integer zeros using :func:`numpy.zeros`:

>>> np.zeros(3, dtype=ComplexInt32DType)
array([(0, 0), (0, 0), (0, 0)], dtype=[('real', '<i2'), ('imag', '<i2')])

Indexing and slicing
--------------------

Indexing the array gives you a complex integer structured scalar:

>>> x = np.array([(1, 2), (3, 4), (5, 6)], dtype=ComplexInt32DType)
>>> x[0]
np.void((1, 2), dtype=[('real', '<i2'), ('imag', '<i2')])
>>> x[1]
np.void((3, 4), dtype=[('real', '<i2'), ('imag', '<i2')])

.. note:
    NumPy displays :any:`numpy.void` because the :any:`ComplexInt32DType` structured data type has
    a base type of :any:`numpy.void`. Using a different base type such as :any:`numpy.int32`
    would have benefits, such as making it easier to convert array elements to/from
    :any:`numpy.int32`, but it would also have drawbacks, such as making it harder to initialize
    the array using a sequence of tuples.

You can index a complex integer structured scalar to get the real and imaginary parts:

>>> x[0][0]
np.int16(1)
>>> x[0][1]
np.int16(2)

You can also index by the field names ``real`` and ``imag``:

>>> x[0]['real']
np.int16(1)
>>> x[0]['imag']
np.int16(2)

Or you can index the entire array by the field names ``real`` and ``imag``:

>>> x['real']
array([1, 3, 5], dtype=int16)
>>> x['imag']
array([2, 4, 6], dtype=int16)

Arrays of complex integers support slicing and negative indices like any other array:

>>> x[0:2]
array([(1, 2), (3, 4)], dtype=[('real', '<i2'), ('imag', '<i2')])
>>> x[1:]
array([(3, 4), (5, 6)], dtype=[('real', '<i2'), ('imag', '<i2')])
>>> x[-1]
np.void((5, 6), dtype=[('real', '<i2'), ('imag', '<i2')])

Conversion
----------

To convert a complex integer structured scalar to a tuple, use the :any:`numpy.ndarray.item`
method:

>>> x[0].item()
(1, 2)
>>> [y.item() for y in x]
[(1, 2), (3, 4), (5, 6)]

To convert NumPy arrays between between different complex number data types, use the
:func:`convert_complex` function:

>>> convert_complex(np.complex128, x)
array([1.+2.j, 3.+4.j, 5.+6.j])
>>> convert_complex(ComplexInt32DType, np.array([1.23+4.56j]))
array([(1, 4)], dtype=[('real', '<i2'), ('imag', '<i2')])

You can also use :func:`convert_complex` with NumPy scalars:

>>> convert_complex(np.complex128, x[0])
np.complex128(1+2j)
>>> convert_complex(ComplexInt32DType, np.complex128(3+4j))
np.void((3, 4), dtype=[('real', '<i2'), ('imag', '<i2')])

.. note::
    As of NumPy 2.2, shape typing is still under development, so its type hints do not reflect that
    many operations coerce zero-dimensional arrays to :any:`numpy.generic`. The type hints for the
    scalar overloads of :func:`convert_complex` follow this precedent and return an
    :any:`numpy.ndarray`. This behavior may change in a future release.

Mathematical operations
-----------------------

Structured arrays of complex integers do not support mathematical operations. Convert
them to arrays of complex floating-point numbers before doing any sort of math or analysis.
"""

from nitypes.complex._conversion import convert_complex
from nitypes.complex._dtypes import ComplexInt32Base, ComplexInt32DType

__all__ = ["convert_complex", "ComplexInt32DType", "ComplexInt32Base"]
__doctest_requires__ = {".": ["numpy>=2.0"]}



from __future__ import annotations

from typing import Any, TypeVar, cast, overload

import numpy as np
import numpy.typing as npt

from nitypes._arguments import validate_dtype
from nitypes._exceptions import unsupported_dtype
from nitypes.complex._dtypes import ComplexInt32DType

_Item_co = TypeVar("_Item_co", bound=Any)
_ScalarType = TypeVar("_ScalarType", bound=np.generic)
_Shape = TypeVar("_Shape", bound=tuple[int, ...])

_COMPLEX_DTYPES = (
    np.complex64,
    np.complex128,
    ComplexInt32DType,
)

_FIELD_DTYPE = {
    np.dtype(np.complex64): np.float32,
    np.dtype(np.complex128): np.float64,
    ComplexInt32DType: np.int16,
}


@overload
def convert_complex(
    requested_dtype: type[_ScalarType] | np.dtype[_ScalarType],
    value: np.ndarray[_Shape, Any],
) -> np.ndarray[_Shape, np.dtype[_ScalarType]]: ...


@overload
def convert_complex(
    requested_dtype: npt.DTypeLike, value: np.ndarray[_Shape, Any]
) -> np.ndarray[_Shape, Any]: ...


# https://numpy.org/doc/2.2/reference/typing.html#d-arrays
# "While thus not strictly correct, all operations are that can potentially perform a 0D-array ->
# scalar cast are currently annotated as exclusively returning an ndarray."
@overload
def convert_complex(
    requested_dtype: type[_ScalarType] | np.dtype[_ScalarType],
    value: np.generic[Any],
) -> np.ndarray[tuple[()], np.dtype[_ScalarType]]: ...


@overload
def convert_complex(
    requested_dtype: npt.DTypeLike,
    value: np.generic[Any],
) -> np.ndarray[tuple[()], Any]: ...


def convert_complex(
    requested_dtype: npt.DTypeLike, value: np.ndarray[_Shape, Any] | np.generic[Any]
) -> np.ndarray[_Shape, Any]:
    """Convert a NumPy array or scalar of complex numbers to the specified dtype.

    Args:
        requested_dtype: The NumPy data type to convert to. Supported data types:
            :any:`numpy.complex64`, :any:`numpy.complex128`, :any:`ComplexInt32DType`.
        value: The NumPy array or scalar to convert.

    Returns:
        The value converted to the specified dtype.
    """
    validate_dtype(requested_dtype, _COMPLEX_DTYPES)
    if requested_dtype == value.dtype:
        return cast(np.ndarray[_Shape, Any], value)
    elif requested_dtype == ComplexInt32DType or value.dtype == ComplexInt32DType:
        # ndarray.view on scalars requires the source and destination types to have the same size,
        # so reshape the scalar into an 1-element array before converting and index it afterwards.
        # shape == () means this is either a scalar (np.generic) or a 0-dimension array, but mypy
        # doesn't know that.
        if value.shape == ():
            return cast(
                np.ndarray[_Shape, Any],
                _convert_complexint32_array(requested_dtype, value.reshape(1))[0],
            )
        else:
            return _convert_complexint32_array(
                requested_dtype, cast(np.ndarray[_Shape, Any], value)
            )
    else:
        return value.astype(requested_dtype)


def _convert_complexint32_array(
    requested_dtype: npt.DTypeLike | type[_ScalarType] | np.dtype[_ScalarType],
    value: np.ndarray[_Shape, Any],
) -> np.ndarray[_Shape, np.dtype[_ScalarType]]:
    if not isinstance(requested_dtype, np.dtype):
        requested_dtype = np.dtype(requested_dtype)

    requested_field_dtype = _FIELD_DTYPE.get(requested_dtype)
    if requested_field_dtype is None:
        raise unsupported_dtype("requested data type", requested_dtype, _COMPLEX_DTYPES)

    value_field_dtype = _FIELD_DTYPE.get(value.dtype)
    if value_field_dtype is None:
        raise unsupported_dtype("array data type", value.dtype, _COMPLEX_DTYPES)

    return value.view(value_field_dtype).astype(requested_field_dtype).view(requested_dtype)



"""Waveform scaling data types for NI Python APIs."""

from nitypes.waveform._scaling._base import ScaleMode
from nitypes.waveform._scaling._linear import LinearScaleMode
from nitypes.waveform._scaling._none import NoneScaleMode

__all__ = ["LinearScaleMode", "NO_SCALING", "NoneScaleMode", "ScaleMode"]

NO_SCALING = NoneScaleMode()
"""A scale mode that does not scale data."""



from __future__ import annotations

from collections.abc import Collection, Iterable, MutableSequence
from typing import (
    TYPE_CHECKING,
    Any,
    final,
    overload,
)

import numpy as np
import numpy.typing as npt

from nitypes._exceptions import invalid_arg_value, invalid_arg_type

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.bintime import CVITimeIntervalDType, TimeDelta, TimeValueTuple
else:
    from nitypes.bintime._dtypes import CVITimeIntervalDType
    from nitypes.bintime._timedelta import TimeDelta
    from nitypes.bintime._time_value_tuple import TimeValueTuple


@final
class TimeDeltaArray(MutableSequence[TimeDelta]):
    """A mutable array of :class:`TimeDelta` values in NI Binary Time Format (NI-BTF).

    Raises:
        TypeError: If any item in value is not a TimeDelta instance.
    """

    __slots__ = ["_array"]

    _array: npt.NDArray[np.void]

    def __init__(
        self,
        value: Collection[TimeDelta] | None = None,
    ) -> None:
        """Initialize a new TimeDeltaArray."""
        if value is None:
            value = []
        if not all(isinstance(item, TimeDelta) for item in value):
            raise invalid_arg_type("value", "iterable of TimeDelta", value)
        self._array = np.fromiter(
            (entry.to_tuple().to_cvi() for entry in value),
            dtype=CVITimeIntervalDType,
            count=len(value),
        )

    @overload
    def __getitem__(  # noqa: D105 - missing docstring in magic method
        self, index: int
    ) -> TimeDelta: ...

    @overload
    def __getitem__(  # noqa: D105 - missing docstring in magic method
        self, index: slice
    ) -> TimeDeltaArray: ...

    def __getitem__(self, index: int | slice) -> TimeDelta | TimeDeltaArray:
        """Return self[index].

        Raises:
            TypeError: If index is an invalid type.
            IndexError: If index is out of range.
        """
        if isinstance(index, int):
            entry = self._array[index].item()
            as_tuple = TimeValueTuple.from_cvi(*entry)
            return TimeDelta.from_tuple(as_tuple)
        elif isinstance(index, slice):
            sliced_entries = self._array[index]
            new_array = TimeDeltaArray()
            new_array._array = sliced_entries
            return new_array
        else:
            raise invalid_arg_type("index", "int or slice", index)

    def __len__(self) -> int:
        """Return len(self)."""
        return len(self._array)

    @overload
    def __setitem__(  # noqa: D105 - missing docstring in magic method
        self, index: int, value: TimeDelta
    ) -> None: ...

    @overload
    def __setitem__(  # noqa: D105 - missing docstring in magic method
        self, index: slice, value: Iterable[TimeDelta]
    ) -> None: ...

    def __setitem__(self, index: int | slice, value: TimeDelta | Iterable[TimeDelta]) -> None:
        """Set a new value for TimeDelta at the specified location or slice.

        Raises:
            TypeError: If index is an invalid type, or slice value is not iterable.
            ValueError: If slice assignment length doesn't match the selected range.
            IndexError: If index is out of range.
        """
        if isinstance(index, int):
            if not isinstance(value, TimeDelta):
                raise invalid_arg_type("value", "TimeDelta", value)
            self._array[index] = value.to_tuple().to_cvi()
        elif isinstance(index, slice):
            if not isinstance(value, Iterable):
                raise invalid_arg_type("value", "iterable of TimeDelta", value)
            if not all(isinstance(item, TimeDelta) for item in value):
                raise invalid_arg_type("value", "iterable of TimeDelta", value)

            start, stop, step = index.indices(len(self))
            selected_count = len(range(start, stop, step))
            values = list(value)
            new_entry_count = len(values)
            if step > 1 and new_entry_count != selected_count:
                raise invalid_arg_value(
                    "value", "iterable with the same length as the slice", value
                )

            if new_entry_count < selected_count:
                # Shrink
                replaced = slice(start, start + new_entry_count)
                removed = slice(start + new_entry_count, stop)
                self._array[replaced] = [item.to_tuple().to_cvi() for item in values]
                del self[removed]
            elif new_entry_count > selected_count:
                # Grow
                replaced = slice(start, stop)
                self._array[replaced] = [
                    item.to_tuple().to_cvi() for item in values[:selected_count]
                ]
                self._array = np.insert(
                    self._array,
                    stop,
                    [item.to_tuple().to_cvi() for item in values[selected_count:]],
                )
            else:
                # Replace, accounting for strides
                self._array[index] = [item.to_tuple().to_cvi() for item in values]
        else:
            raise invalid_arg_type("index", "int or slice", index)

    @overload
    def __delitem__(self, index: int) -> None: ...  # noqa: D105 - missing docstring in magic method

    @overload
    def __delitem__(  # noqa: D105 - missing docstring in magic method
        self, index: slice
    ) -> None: ...

    def __delitem__(self, index: int | slice) -> None:
        """Delete the value at the specified location or slice.

        Raises:
            TypeError: If index is an invalid type.
            IndexError: If index is out of range.
        """
        if isinstance(index, (int, slice)):
            self._array = np.delete(self._array, index)
        else:
            raise invalid_arg_type("index", "int or slice", index)

    def insert(self, index: int, value: TimeDelta) -> None:
        """Insert the TimeDelta value before the specified index.

        Raises:
            TypeError: If index is not int or value is not TimeDelta.
        """
        if not isinstance(index, int):
            raise invalid_arg_type("index", "int", index)
        if not isinstance(value, TimeDelta):
            raise invalid_arg_type("value", "TimeDelta", value)
        lower = -len(self._array)
        upper = len(self._array)
        index = min(max(index, lower), upper)
        as_cvi = value.to_tuple().to_cvi()
        self._array = np.insert(self._array, index, as_cvi)

    def __eq__(self, other: object) -> bool:
        """Return self == other."""
        if not isinstance(other, TimeDeltaArray):
            return NotImplemented
        return np.array_equal(self._array, other._array)

    def __reduce__(self) -> tuple[Any, ...]:
        """Return object state for pickling."""
        return (self.__class__, (list(iter(self)),))

    def __repr__(self) -> str:
        """Return repr(self)."""
        ctor_args = list(iter(self))
        return f"{self.__class__.__module__}.{self.__class__.__name__}({ctor_args})"

    def __str__(self) -> str:
        """Return str(self)."""
        values = list(iter(self))
        return f"[{'; '.join(str(v) for v in values)}]"



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy.typing as npt

from nitypes.waveform._scaling._base import _ScalarType

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.waveform import ScaleMode
else:
    from nitypes.waveform._scaling._base import ScaleMode


class NoneScaleMode(ScaleMode):
    """A scale mode that does not scale data."""

    __slots__ = ()

    def _transform_data(self, data: npt.NDArray[_ScalarType]) -> npt.NDArray[_ScalarType]:
        return data

    def __eq__(self, value: object, /) -> bool:
        """Return self==value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        return True

    def __repr__(
        self,
    ) -> str:
        """Return repr(self)."""
        return f"{self.__class__.__module__}.{self.__class__.__name__}()"



from __future__ import annotations

import numpy as np
from typing_extensions import TypeAlias

ComplexInt32Base: TypeAlias = np.void
"""Type alias for the base type of :any:`ComplexInt32DType`, which is :any:`numpy.void`."""

ComplexInt32DType = np.dtype((ComplexInt32Base, [("real", np.int16), ("imag", np.int16)]))
"""NumPy structured data type for a complex integer with 16-bit ``real`` and ``imag`` fields."""



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Generic

from nitypes.waveform._timing._sample_interval._mode import SampleIntervalMode
from nitypes.waveform.typing import TSampleInterval_co, TTimeOffset_co, TTimestamp_co

if TYPE_CHECKING:
    from nitypes.waveform._timing._timing import Timing  # circular import


class SampleIntervalStrategy(ABC, Generic[TTimestamp_co, TTimeOffset_co, TSampleInterval_co]):
    """Implements SampleIntervalMode specific behavior."""

    # Note that timing is always passed as a parameter. The timing object has a reference to the
    # strategy, so saving a reference to the timing object would introduce a reference cycle.
    __slots__ = ()

    @abstractmethod
    def validate_init_args(
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        sample_interval_mode: SampleIntervalMode,
        timestamp: TTimestamp_co | None,
        time_offset: TTimeOffset_co | None,
        sample_interval: TSampleInterval_co | None,
        timestamps: Sequence[TTimestamp_co] | None,
    ) -> None:
        """Validate the BaseTiming.__init__ arguments for this mode."""
        raise NotImplementedError

    @abstractmethod
    def get_timestamps(
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        start_index: int,
        count: int,
    ) -> Iterable[TTimestamp_co]:
        """Get or generate timestamps for the specified samples."""
        raise NotImplementedError

    @abstractmethod
    def append_timestamps(
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        timestamps: Sequence[TTimestamp_co] | None,
    ) -> Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co]:
        """Append timestamps and return a new waveform timing if needed."""
        raise NotImplementedError

    @abstractmethod
    def append_timing(
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        other: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
    ) -> Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co]:
        """Append timing and return a new waveform timing if needed."""
        raise NotImplementedError



"""Scalar data types for NI Python APIs.

Scalar Data Type
=================

:class:`Scalar`: A scalar data object represents a single scalar value with units information.
Valid types for the scalar value are :any:`bool`, :any:`int`, :any:`float`, and :any:`str`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Union

from typing_extensions import TypeVar, final

from nitypes._exceptions import invalid_arg_type, invalid_arg_value
from nitypes.waveform._extended_properties import UNIT_DESCRIPTION

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.waveform import ExtendedPropertyDictionary
else:
    from nitypes.waveform._extended_properties import ExtendedPropertyDictionary

TScalar_co = TypeVar("TScalar_co", bound=Union[bool, int, float, str], covariant=True)
_NUMERIC = (bool, int, float)


@final
class Scalar(Generic[TScalar_co]):
    """A scalar data class, which encapsulates scalar data and units information.

    Constructing
    ^^^^^^^^^^^^

    To construct a scalar data object, use the :class:`Scalar` class:

    >>> Scalar(False)
    nitypes.scalar.Scalar(value=False, units='')
    >>> Scalar(0)
    nitypes.scalar.Scalar(value=0, units='')
    >>> Scalar(5.0, 'volts')
    nitypes.scalar.Scalar(value=5.0, units='volts')
    >>> Scalar("value", "volts")
    nitypes.scalar.Scalar(value='value', units='volts')

    Class members
    ^^^^^^^^^^^^^
    """

    __slots__ = [
        "_value",
        "_extended_properties",
    ]

    _value: TScalar_co
    _extended_properties: ExtendedPropertyDictionary

    def __init__(
        self,
        value: TScalar_co,
        units: str = "",
    ) -> None:
        """Initialize a new scalar.

        Args:
            value: The scalar data to store in this object.
            units: The units string associated with this data.

        Returns:
            A scalar data object.
        """
        if not isinstance(value, (bool, int, float, str)):
            raise invalid_arg_type("scalar input data", "bool, int, float, or str", value)

        if not isinstance(units, str):
            raise invalid_arg_type("units", "str", units)

        # The ScalarData proto type only supports 32 bit integers. Make sure we're in range.
        if isinstance(value, int):
            if value <= -0x80000000 or value >= 0x7FFFFFFF:
                raise invalid_arg_value("integer scalar value", "within the range of Int32", value)

        self._value = value
        self._extended_properties = ExtendedPropertyDictionary()
        self._extended_properties[UNIT_DESCRIPTION] = units

    @property
    def value(self) -> TScalar_co:
        """The scalar value."""
        return self._value

    @property
    def units(self) -> str:
        """The unit of measurement, such as volts, of the scalar."""
        value = self._extended_properties.get(UNIT_DESCRIPTION, "")
        assert isinstance(value, str)
        return value

    @units.setter
    def units(self, value: str) -> None:
        if not isinstance(value, str):
            raise invalid_arg_type("units", "str", value)
        self._extended_properties[UNIT_DESCRIPTION] = value

    @property
    def extended_properties(self) -> ExtendedPropertyDictionary:
        """The extended properties for the scalar.

        .. note::
            Data stored in the extended properties dictionary may not be encrypted when you send it
            over the network or write it to a TDMS file.
        """
        return self._extended_properties

    def __eq__(self, value: object, /) -> bool:
        """Return self==value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        return self.value == value.value and self.units == value.units

    def __gt__(self, value: Scalar[TScalar_co]) -> bool:
        """Return self > value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        self._check_units_equal_for_comparison(value.units)
        if isinstance(self.value, _NUMERIC) and isinstance(value.value, _NUMERIC):
            return self.value > value.value  # type: ignore[no-any-return,operator]  # https://github.com/python/mypy/issues/19454
        elif isinstance(self.value, str) and isinstance(value.value, str):
            return self.value > value.value
        else:
            raise TypeError("Comparing Scalar objects of numeric and string types is not permitted")

    def __ge__(self, value: Scalar[TScalar_co]) -> bool:
        """Return self >= value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        self._check_units_equal_for_comparison(value.units)
        if isinstance(self.value, _NUMERIC) and isinstance(value.value, _NUMERIC):
            return self.value >= value.value  # type: ignore[no-any-return,operator]  # https://github.com/python/mypy/issues/19454
        elif isinstance(self.value, str) and isinstance(value.value, str):
            return self.value >= value.value
        else:
            raise TypeError("Comparing Scalar objects of numeric and string types is not permitted")

    def __lt__(self, value: Scalar[TScalar_co]) -> bool:
        """Return self < value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        self._check_units_equal_for_comparison(value.units)
        if isinstance(self.value, _NUMERIC) and isinstance(value.value, _NUMERIC):
            return self.value < value.value  # type: ignore[no-any-return,operator]  # https://github.com/python/mypy/issues/19454
        elif isinstance(self.value, str) and isinstance(value.value, str):
            return self.value < value.value
        else:
            raise TypeError("Comparing Scalar objects of numeric and string types is not permitted")

    def __le__(self, value: Scalar[TScalar_co]) -> bool:
        """Return self <= value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        self._check_units_equal_for_comparison(value.units)
        if isinstance(self.value, _NUMERIC) and isinstance(value.value, _NUMERIC):
            return self.value <= value.value  # type: ignore[no-any-return,operator]  # https://github.com/python/mypy/issues/19454
        elif isinstance(self.value, str) and isinstance(value.value, str):
            return self.value <= value.value
        else:
            raise TypeError("Comparing Scalar objects of numeric and string types is not permitted")

    def __reduce__(self) -> tuple[Any, ...]:
        """Return object state for pickling."""
        return (self.__class__, (self.value, self.units))

    def __repr__(self) -> str:
        """Return repr(self)."""
        args = [f"value={self.value!r}", f"units={self.units!r}"]
        return f"{self.__class__.__module__}.{self.__class__.__name__}({', '.join(args)})"

    def __str__(self) -> str:
        """Return str(self)."""
        value_str = str(self.value)
        if self.units:
            value_str += f" {self.units}"

        return value_str

    def _check_units_equal_for_comparison(self, other_units: str) -> None:
        """Raise a ValueError if other_units != self.units."""
        if self.units != other_units:
            raise ValueError("Comparing Scalar objects with different units is not permitted.")



from __future__ import annotations

from typing import TYPE_CHECKING, SupportsFloat

import numpy.typing as npt

from nitypes._arguments import arg_to_float
from nitypes.waveform._scaling._base import _ScalarType

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.waveform import ScaleMode
else:
    from nitypes.waveform._scaling._base import ScaleMode


class LinearScaleMode(ScaleMode):
    """A scale mode that scales data linearly."""

    __slots__ = ["_gain", "_offset", "__weakref__"]

    _gain: float
    _offset: float

    def __init__(self, gain: SupportsFloat, offset: SupportsFloat) -> None:
        """Initialize a new scale mode object that scales data linearly.

        Args:
            gain: The gain of the linear scale.
            offset: The offset of the linear scale.

        Returns:
            A scale mode that scales data linearly.
        """
        self._gain = arg_to_float("gain", gain)
        self._offset = arg_to_float("offset", offset)

    @property
    def gain(self) -> float:
        """The gain of the linear scale."""
        return self._gain

    @property
    def offset(self) -> float:
        """The offset of the linear scale."""
        return self._offset

    def _transform_data(self, data: npt.NDArray[_ScalarType]) -> npt.NDArray[_ScalarType]:
        # https://github.com/numpy/numpy/issues/28805 - TYP: mypy infers that adding/multiplying a
        # npt.NDArray[np.float32] with a float promotes dtype to Any or np.float64
        return data * self._gain + self._offset  # type: ignore[operator,no-any-return]

    def __eq__(self, value: object, /) -> bool:
        """Return self==value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        return self._gain == value._gain and self._offset == value._offset

    def __repr__(
        self,
    ) -> str:
        """Return repr(self)."""
        return f"{self.__class__.__module__}.{self.__class__.__name__}({self.gain}, {self.offset})"



from __future__ import annotations

from enum import IntEnum

from nitypes._exceptions import invalid_arg_value

_CHAR_TABLE = "01ZLHXTV"

_STATE_TEST_TABLE = [
    # 0  1  Z  L  H  X  T  V
    [1, 0, 0, 1, 0, 1, 0, 1],  # 0
    [0, 1, 0, 0, 1, 1, 0, 1],  # 1
    [0, 0, 1, 0, 0, 1, 1, 0],  # Z
    [1, 0, 0, 1, 0, 1, 0, 0],  # L
    [0, 1, 0, 0, 1, 1, 0, 0],  # H
    [1, 1, 1, 1, 1, 1, 1, 1],  # X
    [0, 0, 1, 0, 0, 1, 1, 0],  # T
    [1, 1, 0, 0, 0, 1, 0, 1],  # V
]


class DigitalState(IntEnum):
    """An IntEnum of the different digital states that a digital signal can represent.

    You can use :class:`DigitalState` in place of an :any:`int`:

    >>> DigitalState.FORCE_OFF
    <DigitalState.FORCE_OFF: 2>
    >>> DigitalState.FORCE_OFF == 2
    True

    Use :meth:`from_char` and :meth:`to_char` to convert between states and characters:

    >>> DigitalState.from_char("Z")
    <DigitalState.FORCE_OFF: 2>
    >>> DigitalState.to_char(2)
    'Z'

    Use :meth:`test` to compare actual vs. expected states, returning True on failure.

    >>> DigitalState.test(DigitalState.FORCE_DOWN, DigitalState.COMPARE_LOW)
    False
    >>> DigitalState.test(DigitalState.FORCE_UP, DigitalState.COMPARE_LOW)
    True
    """

    _value_: int

    FORCE_DOWN = 0
    """Force logic low (``0``). Drive to the low voltage level (VIL)."""

    FORCE_UP = 1
    """Force logic high (``1``). Drive to the high voltage level (VIH)."""

    FORCE_OFF = 2
    """Force logic high impedance (``Z``). Turn the driver off."""

    COMPARE_LOW = 3
    """Compare logic low (edge) (``L``). Compare for a voltage level lower than the low voltage
    threshold (VOL)."""

    COMPARE_HIGH = 4
    """Compare logic high (edge) (``H``). Compare for a voltage level higher than the high voltage
    threshold (VOH)."""

    COMPARE_UNKNOWN = 5
    """Compare logic unknown (``X``). Don't compare."""

    COMPARE_OFF = 6
    """Compare logic high impedance (edge) (``T``). Compare for a voltage level between the low
    voltage threshold (VOL) and the high voltage threshold (VOH)."""

    COMPARE_VALID = 7
    """Compare logic valid level (edge) (``V``). Compare for a voltage level either lower than the
    low voltage threshold (VOL) or higher than the high voltage threshold (VOH)."""

    @property
    def char(self) -> str:
        """The character representing the digital state."""
        return _CHAR_TABLE[self]

    @classmethod
    def from_char(cls, char: str) -> DigitalState:
        """Look up the digital state for the corresponding character."""
        try:
            return DigitalState(_CHAR_TABLE.index(char))
        except ValueError:
            raise KeyError(char)

    @classmethod
    def to_char(cls, state: DigitalState, errors: str = "strict") -> str:
        """Get a character representing the digital state.

        Args:
            state: The digital state.
            errors: Specifies how to handle errors.

                * "strict": raise ``KeyError``
                * "replace": return "?"

        Returns:
            A character representing the digital state.
        """
        if errors not in ("strict", "replace"):
            raise invalid_arg_value("errors argument", "supported value", errors)
        try:
            return DigitalState(state).char
        except ValueError:
            if errors == "strict":
                raise KeyError(state)
            elif errors == "replace":
                return "?"
            raise

    @staticmethod
    def test(state1: DigitalState, state2: DigitalState) -> bool:
        """Test two digital states and return True if the test failed."""
        return not _STATE_TEST_TABLE[DigitalState(state1)][DigitalState(state2)]



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Generic, overload

from nitypes._exceptions import invalid_arg_type
from nitypes.waveform.typing import TDigitalState

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.waveform import DigitalWaveform, DigitalWaveformSignal
else:
    # DigitalWaveform is a circular import.
    from nitypes.waveform._digital._signal import DigitalWaveformSignal


class DigitalWaveformSignalCollection(
    Generic[TDigitalState], Sequence[DigitalWaveformSignal[TDigitalState]]
):
    """A collection of digital waveform signals.

    To construct this object, use the :any:`DigitalWaveform.signals` property.
    """

    __slots__ = ["_owner", "_signals", "__weakref__"]

    _owner: DigitalWaveform[TDigitalState]
    _signals: list[DigitalWaveformSignal[TDigitalState] | None]

    def __init__(self, owner: DigitalWaveform[TDigitalState]) -> None:
        """Initialize a new DigitalWaveformSignalCollection."""
        self._owner = owner
        self._signals = [None] * owner.signal_count

    def __len__(self) -> int:
        """Return len(self)."""
        return len(self._signals)

    @overload
    def __getitem__(  # noqa: D105 - missing docstring in magic method
        self, index: int | str
    ) -> DigitalWaveformSignal[TDigitalState]: ...
    @overload
    def __getitem__(  # noqa: D105 - missing docstring in magic method
        self, index: slice
    ) -> Sequence[DigitalWaveformSignal[TDigitalState]]: ...

    def __getitem__(
        self, index: int | str | slice
    ) -> DigitalWaveformSignal[TDigitalState] | Sequence[DigitalWaveformSignal[TDigitalState]]:
        """Get self[index]."""
        if isinstance(index, int):
            if index < 0:
                index += len(self._signals)
            value = self._signals[index]
            if value is None:
                value = self._signals[index] = DigitalWaveformSignal(self._owner, index)
            return value
        elif isinstance(index, str):
            signal_names = self._owner._get_signal_names()
            try:
                signal_index = signal_names.index(index)
            except ValueError:
                raise IndexError(index)
            return self[signal_index]
        elif isinstance(index, slice):
            return [self[i] for i in range(*index.indices(len(self)))]
        else:
            raise invalid_arg_type("index", "int or str", index)



from __future__ import annotations

import reprlib
import sys

import numpy as np
import numpy.typing as npt


def add_note(exception: Exception, note: str) -> None:
    """Add a note to an exception.

    >>> try:
    ...     raise ValueError("Oh no")
    ... except Exception as e:
    ...     add_note(e, "p.s. This is bad")
    ...     raise
    Traceback (most recent call last):
    ...
    ValueError: Oh no
    p.s. This is bad
    """
    if sys.version_info >= (3, 11):
        exception.add_note(note)
    else:
        message = exception.args[0] + "\n" + note
        exception.args = (message,) + exception.args[1:]


def invalid_arg_value(
    arg_description: str, valid_value_description: str, value: object
) -> ValueError:
    """Create a ValueError for an invalid argument value."""
    return ValueError(
        f"The {arg_description} must be {_a(valid_value_description)}.\n\n"
        f"Provided value: {reprlib.repr(value)}"
    )


def invalid_arg_type(arg_description: str, type_description: str, value: object) -> TypeError:
    """Create a TypeError for an invalid argument type."""
    return TypeError(
        f"The {arg_description} must be {_a(type_description)}.\n\n"
        f"Provided value: {reprlib.repr(value)}"
    )


def invalid_array_ndim(arg_description: str, valid_value_description: str, ndim: int) -> ValueError:
    """Create a ValueError for an array with an invalid number of dimensions."""
    return ValueError(
        f"The {arg_description} must be {_a(valid_value_description)}.\n\n"
        f"Number of dimensions: {ndim}"
    )


def invalid_requested_type(type_description: str, requested_type: type) -> TypeError:
    """Create a TypeError for an invalid requested type."""
    return TypeError(
        f"The requested type must be {_a(type_description)} type.\n\n"
        f"Requested type: {requested_type}"
    )


def unsupported_arg(arg_description: str, value: object) -> ValueError:
    """Create a ValueError for an unsupported argument."""
    return ValueError(
        f"The {arg_description} argument is not supported.\n\n"
        f"Provided value: {reprlib.repr(value)}"
    )


def unsupported_dtype(
    arg_description: str, dtype: npt.DTypeLike, supported_dtypes: tuple[npt.DTypeLike, ...]
) -> TypeError:
    """Create a TypeError for an unsupported dtype."""
    # Remove duplicate names because distinct types (e.g. int vs. long) may have the same name
    # ("int32").
    supported_dtype_names = {np.dtype(d).name: None for d in supported_dtypes}.keys()
    return TypeError(
        f"The {arg_description} is not supported.\n\n"
        f"Data type: {np.dtype(dtype)}\n"
        f"Supported data types: {', '.join(supported_dtype_names)}"
    )


def int_out_of_range(value: int, min: int, max: int) -> OverflowError:
    """Create an OverflowError when an int is out of the specified range."""
    raise OverflowError(
        "The input value is out of range.\n\n"
        f"Requested value: {value}\n"
        f"Minimum value: {min}\n",
        f"Maximum value: {max}",
    )


# English-specific hack. This is why we prefer "Key: value" for localizable errors. TODO: consider
# moving the full strings into a string table instead of building them out of English noun phrases.
def _a(noun: str) -> str:
    indefinite_article = "an" if noun[0] in "AEIOUaeiou" else "a"
    if noun.startswith("one-"):
        indefinite_article = "a"
    return f"{indefinite_article} {noun}"



"""Waveform data types for NI Python APIs.

Waveform Data Types
===================

* :class:`AnalogWaveform`: An analog waveform represents a single analog signal with timing
  information and extended properties such as units.
* :class:`ComplexWaveform`: A complex waveform represents a single complex-number signal, such as
  I/Q data, with timing information and extended properties such as units.
* :class:`DigitalWaveform`: A digital waveform represents one or more digital signals with timing
  information and extended properties such as channel name and signal names.
* :class:`Spectrum`: A frequency spectrum represents an analog signal with frequency information
  and extended properties such as units.
"""

from nitypes.waveform._analog import AnalogWaveform
from nitypes.waveform._complex import ComplexWaveform
from nitypes.waveform._digital import (
    DigitalState,
    DigitalWaveform,
    DigitalWaveformFailure,
    DigitalWaveformSignal,
    DigitalWaveformSignalCollection,
    DigitalWaveformTestResult,
)
from nitypes.waveform._extended_properties import ExtendedPropertyDictionary
from nitypes.waveform._numeric import NumericWaveform
from nitypes.waveform._scaling import (
    NO_SCALING,
    LinearScaleMode,
    NoneScaleMode,
    ScaleMode,
)
from nitypes.waveform._spectrum import Spectrum
from nitypes.waveform._timing import SampleIntervalMode, Timing

__all__ = [
    "AnalogWaveform",
    "ComplexWaveform",
    "DigitalState",
    "DigitalWaveform",
    "DigitalWaveformFailure",
    "DigitalWaveformSignal",
    "DigitalWaveformSignalCollection",
    "DigitalWaveformTestResult",
    "ExtendedPropertyDictionary",
    "LinearScaleMode",
    "NO_SCALING",
    "NoneScaleMode",
    "NumericWaveform",
    "SampleIntervalMode",
    "ScaleMode",
    "Spectrum",
    "Timing",
]


# Hide that it was defined in a helper file
AnalogWaveform.__module__ = __name__
ComplexWaveform.__module__ = __name__
DigitalState.__module__ = __name__
DigitalWaveform.__module__ = __name__
DigitalWaveformFailure.__module__ = __name__
DigitalWaveformSignal.__module__ = __name__
DigitalWaveformSignalCollection.__module__ = __name__
DigitalWaveformTestResult.__module__ = __name__
ExtendedPropertyDictionary.__module__ = __name__
LinearScaleMode.__module__ = __name__
# NO_SCALING is a constant
NoneScaleMode.__module__ = __name__
NumericWaveform.__module__ = __name__
SampleIntervalMode.__module__ = __name__
ScaleMode.__module__ = __name__
Spectrum.__module__ = __name__
Timing.__module__ = __name__



"""Type aliases and type variables for waveforms."""

# These types are in a submodule so they don't show up in autocomplete for "nitypes.waveform." and
# overwhelm users.

from __future__ import annotations

import datetime as dt
from typing import Union

import numpy as np
from typing_extensions import TypeAlias, TypeVar

from nitypes._numpy import bool as _np_bool
from nitypes.time.typing import AnyDateTime, AnyTimeDelta

ExtendedPropertyValue: TypeAlias = Union[bool, float, int, str]
"""Type alias for an ExtendedPropertyDictionary value.

This type alias is a union of the following types:
* :class:`bool`
* :class:`float`
* :class:`int`
* :class:`str`
"""

AnyDigitalPort: TypeAlias = Union[np.uint8, np.uint16, np.uint32]
"""Type alias for any digital port data type.

This type alias is a union of the following types:

* :class:`numpy.uint8`
* :class:`numpy.uint16`
* :class:`numpy.uint32`
"""

# np.byte == np.int8, np.ubyte == np.uint8
AnyDigitalState: TypeAlias = Union[_np_bool, np.int8, np.uint8]
"""Type alias for any digital state data type.

This type alias is a union of the following types:

* :class:`numpy.bool` (NumPy 2.x) or :class:`numpy.bool_` (NumPy 1.x)
* :class:`numpy.int8`
* :class:`numpy.uint8`
"""

TDigitalState = TypeVar("TDigitalState", bound=AnyDigitalState)
"""Type variable with a bound of :any:`AnyDigitalState`."""

TOtherDigitalState = TypeVar("TOtherDigitalState", bound=AnyDigitalState)
"""Another type variable with a bound of :any:`AnyDigitalState`."""

TTimestamp = TypeVar("TTimestamp", bound=AnyDateTime, default=dt.datetime)
"""Type variable for a timestamp."""

TTimestamp_co = TypeVar(
    "TTimestamp_co",
    bound=AnyDateTime,
    covariant=True,
    default=dt.datetime,
)
"""Covariant type variable for a timestamp."""

TTimeOffset = TypeVar(
    "TTimeOffset",
    bound=AnyTimeDelta,
    default=dt.timedelta,
)
"""Type variable for a time offset."""

TTimeOffset_co = TypeVar(
    "TTimeOffset_co",
    bound=AnyTimeDelta,
    covariant=True,
    default=dt.timedelta,
)
"""Covariant type variable for a time offset."""

TSampleInterval = TypeVar(
    "TSampleInterval",
    bound=AnyTimeDelta,
    default=dt.timedelta,
)
"""Type variable for a sample interval."""

TSampleInterval_co = TypeVar(
    "TSampleInterval_co",
    bound=AnyTimeDelta,
    covariant=True,
    default=dt.timedelta,
)
"""Covariant type variable for a sample interval."""

TOtherTimestamp = TypeVar("TOtherTimestamp", bound=AnyDateTime)
"""Another type variable for a timestamp."""

TOtherTimeOffset = TypeVar("TOtherTimeOffset", bound=AnyTimeDelta)
"""Another type variable for a time offset."""

TOtherSampleInterval = TypeVar("TOtherSampleInterval", bound=AnyTimeDelta)
"""Another type variable for a sample interval."""



"""Custom error classes for waveforms."""

from __future__ import annotations


class TimingMismatchError(RuntimeError):
    """Exception used when appending waveforms with mismatched timing."""

    pass


class CapacityMismatchError(ValueError):
    """An error for an invalid capacity."""

    pass


class CapacityTooSmallError(ValueError):
    """An error for an invalid capacity argument."""

    pass


class DatatypeMismatchError(TypeError):
    """An error for a data type mismatch."""

    pass


class IrregularTimestampCountMismatchError(ValueError):
    """An error for an irregular timestamp count mismatch."""

    pass


class StartIndexTooLargeError(ValueError):
    """An error for an invalid start index argument."""

    pass


class StartIndexOrSampleCountTooLargeError(ValueError):
    """An error for an invalid start index or sample count argument."""

    pass


class NoTimestampInformationError(RuntimeError):
    """An error for waveform timing with no timestamp information."""

    pass


class SampleIntervalModeMismatchError(TimingMismatchError):
    """An error for mixing none/regular with irregular timing."""

    pass


class SignalCountMismatchError(ValueError):
    """An error for a mismatched signal count."""

    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, SupportsIndex

import numpy as np
import numpy.typing as npt

from nitypes._arguments import arg_to_uint
from nitypes.waveform.typing import TDigitalState

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.waveform import DigitalWaveform
else:
    # DigitalWaveform is a circular import.
    pass


class DigitalWaveformSignal(Generic[TDigitalState]):
    """A signal of a digital waveform.

    To construct this object, use the :any:`DigitalWaveform.signals` property and index the returned
    collection, e.g. ``waveform.signals[0]`` or ``waveform.signals["Dev1/port0/line0"]``.
    """

    __slots__ = ["_owner", "_signal_index", "__weakref__"]

    _owner: DigitalWaveform[TDigitalState]
    _signal_index: int

    def __init__(self, owner: DigitalWaveform[TDigitalState], signal_index: SupportsIndex) -> None:
        """Initialize a new digital waveform signal."""
        self._owner = owner
        self._signal_index = arg_to_uint("signal index", signal_index)

    @property
    def owner(self) -> DigitalWaveform[TDigitalState]:
        """The waveform that owns this signal."""
        return self._owner

    @property
    def signal_index(self) -> int:
        """The signal index."""
        return self._signal_index

    @property
    def data(self) -> npt.NDArray[TDigitalState]:
        """The signal data, indexed by sample."""
        return self._owner.data[:, self._signal_index]

    @property
    def name(self) -> str:
        """The signal name."""
        return self._owner._get_signal_name(self._signal_index)

    @name.setter
    def name(self, value: str) -> None:
        self._owner._set_signal_name(self._signal_index, value)

    def __eq__(self, value: object, /) -> bool:
        """Return self==value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        # Do not compare the index or name.
        return np.array_equal(self.data, value.data)

    def __reduce__(self) -> tuple[Any, ...]:
        """Return object state for pickling."""
        ctor_args = (self._owner, self._signal_index)
        return (self.__class__, ctor_args)

    def __repr__(self) -> str:
        """Return repr(self)."""
        # This is not the same as the constructor arguments.
        args = []
        if self.name:
            args.append(f"name={self.name!r}")
        if self._owner._sample_count > 0:
            args.append(f"data={self.data!r}")
        return f"{self.__class__.__module__}.{self.__class__.__name__}({', '.join(args)})"



from __future__ import annotations

import datetime as dt
from functools import singledispatchmethod
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    SupportsIndex,
    Union,
    cast,
    final,
    overload,
)

import hightime as ht
from typing_extensions import Self, TypeAlias

from nitypes._exceptions import invalid_arg_type, invalid_arg_value
from nitypes.bintime._timedelta import (
    _OTHER_TIMEDELTA_TUPLE,
    _OtherTimeDelta,
)

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.bintime import TimeDelta, TimeValueTuple
else:
    from nitypes.bintime._time_value_tuple import TimeValueTuple
    from nitypes.bintime._timedelta import TimeDelta

_DT_EPOCH_1904 = dt.datetime(1904, 1, 1, tzinfo=dt.timezone.utc)
_HT_EPOCH_1904 = ht.datetime(1904, 1, 1, tzinfo=dt.timezone.utc)

_OtherDateTime: TypeAlias = Union[dt.datetime, ht.datetime]
_OTHER_DATETIME_TUPLE = (dt.datetime, ht.datetime)


@final
class DateTime:
    """An absolute time in NI Binary Time Format (NI-BTF).

    DateTime represents time as a 128-bit fixed point number with 64-bit whole seconds and
    64-bit fractional seconds.

    .. warning::
        The fractional seconds are represented as a binary fraction, which is a sum of inverse
        powers of 2. Values that are not exactly representable as binary fractions will display
        rounding error or "bruising" similar to a floating point number.

    DateTime instances are duck typing compatible with a subset of the method and properties
    supported by :any:`datetime.datetime` and :any:`hightime.datetime`.

    This class only supports the UTC time zone and does not support timezone-naive times.

    This class does not support the ``fold`` property for disambiguating repeated times for daylight
    saving time and time zone changes.

    Constructing
    ^^^^^^^^^^^^

    As with :any:`datetime.datetime`, you can construct a :class:`DateTime` by specifying the year,
    month, day, etc.:

    >>> import datetime
    >>> DateTime(2025, 5, 25, 16, 45, tzinfo=datetime.timezone.utc)
    nitypes.bintime.DateTime(2025, 5, 25, 16, 45, tzinfo=datetime.timezone.utc)

    .. note::
        :class:`DateTime` only supports :any:`datetime.timezone.utc`. It does not support time-zone-naive
        objects or time zones other than UTC.

    You can also construct a :class:`DateTime` from a :any:`datetime.datetime` or
    :any:`hightime.datetime`:

    >>> DateTime(datetime.datetime(2025, 5, 25, 16, 45, tzinfo=datetime.timezone.utc))
    nitypes.bintime.DateTime(2025, 5, 25, 16, 45, tzinfo=datetime.timezone.utc)
    >>> import hightime
    >>> DateTime(hightime.datetime(2025, 5, 25, 16, 45, tzinfo=datetime.timezone.utc))
    nitypes.bintime.DateTime(2025, 5, 25, 16, 45, tzinfo=datetime.timezone.utc)

    You can get the current time of day by calling :any:`DateTime.now`:

    >>> DateTime.now(datetime.timezone.utc) # doctest: +ELLIPSIS
    nitypes.bintime.DateTime(...)

    Properties
    ^^^^^^^^^^

    Like other ``datetime`` objects, :class:`DateTime` has properties for the year, month, day, hour,
    minute, second, and microsecond.

    >>> import datetime
    >>> x = DateTime(datetime.datetime(2025, 5, 25, 16, 45, tzinfo=datetime.timezone.utc))
    >>> (x.year, x.month, x.day)
    (2025, 5, 25)
    >>> (x.hour, x.minute, x.second, x.microsecond)
    (16, 45, 0, 0)

    Like :any:`hightime.datetime`, it also supports the femtosecond and yoctosecond properties.

    >>> (x.femtosecond, x.yoctosecond)
    (0, 0)

    Resolution
    ^^^^^^^^^^

    NI-BTF is a high-resolution time format, so it has significantly higher resolution than
    :any:`datetime.datetime`. However, :any:`hightime.datetime` has even higher resolution:

    ========================   ================================
    Class                      Smallest Time Increment
    ========================   ================================
    :any:`datetime.datetime`   1 microsecond (1e-6 sec)
    :class:`DateTime`            54210 yoctoseconds (5.4e-20 sec)
    :any:`hightime.datetime`   1 yoctosecond (1e-24 sec)
    ========================   ================================

    As a result, :any:`hightime.datetime` can represent the time down to the exact yoctosecond, but
    :class:`DateTime` rounds the yoctosecond field.

    >>> import hightime
    >>> x = hightime.datetime(2025, 1, 1, yoctosecond=123456789, tzinfo=datetime.timezone.utc)
    >>> x
    hightime.datetime(2025, 1, 1, 0, 0, 0, 0, 0, 123456789, tzinfo=datetime.timezone.utc)
    >>> DateTime(x) # doctest: +NORMALIZE_WHITESPACE
    nitypes.bintime.DateTime(2025, 1, 1, 0, 0, 0, 0, 0, 123436417, tzinfo=datetime.timezone.utc)

    Rounding
    ^^^^^^^^

    NI-BTF represents fractional seconds as a binary fraction, which is a sum of inverse
    powers of 2. Values that are not exactly representable as binary fractions will display
    rounding error or "bruising" similar to a floating point number.

    For example, it may round 100 microseconds down to 99.9999... microseconds.

    >>> x = hightime.datetime(2025, 1, 1, microsecond=100, tzinfo=datetime.timezone.utc)
    >>> x
    hightime.datetime(2025, 1, 1, 0, 0, 0, 100, tzinfo=datetime.timezone.utc)
    >>> DateTime(x) # doctest: +NORMALIZE_WHITESPACE
    nitypes.bintime.DateTime(2025, 1, 1, 0, 0, 0, 99, 999999999, 999991239,
        tzinfo=datetime.timezone.utc)

    Class members
    ^^^^^^^^^^^^^
    """  # noqa: W505 - doc line too long

    min: ClassVar[DateTime]
    """The earliest supported :class:`DateTime` object, midnight on Jan 1, 0001, UTC."""

    max: ClassVar[DateTime]
    """The latest supported :class:`DateTime` object, before midnight on Dec 31, 9999, UTC."""

    __slots__ = ["_offset", "_hightime_cache"]

    _offset: TimeDelta
    _hightime_cache: ht.datetime | None

    @overload
    def __init__(self) -> None: ...  # noqa: D107 - missing docstring in __init__

    @overload
    def __init__(  # noqa: D107 - missing docstring in __init__
        self, value: _OtherDateTime, /
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - missing docstring in __init__
        self,
        year: SupportsIndex,
        month: SupportsIndex,
        day: SupportsIndex,
        hour: SupportsIndex = ...,
        minute: SupportsIndex = ...,
        second: SupportsIndex = ...,
        microsecond: SupportsIndex = ...,
        femtosecond: SupportsIndex = ...,
        yoctosecond: SupportsIndex = ...,
        tzinfo: dt.tzinfo | None = None,
    ) -> None: ...

    def __init__(
        self,
        year: SupportsIndex | _OtherDateTime | None = None,
        month: SupportsIndex | None = None,
        day: SupportsIndex | None = None,
        hour: SupportsIndex = 0,
        minute: SupportsIndex = 0,
        second: SupportsIndex = 0,
        microsecond: SupportsIndex = 0,
        femtosecond: SupportsIndex = 0,
        yoctosecond: SupportsIndex = 0,
        tzinfo: dt.tzinfo | None = None,
    ) -> None:
        """Initialize a DateTime."""
        if isinstance(year, SupportsIndex):
            self._offset = self.__class__._to_offset(
                ht.datetime(
                    year,
                    cast(SupportsIndex, month),
                    cast(SupportsIndex, day),
                    hour,
                    minute,
                    second,
                    microsecond,
                    femtosecond,
                    yoctosecond,
                    tzinfo,
                )
            )
        else:
            self._offset = self.__class__._to_offset(year)

        # Do not cache the passed-in ht.datetime because that would hide the rounding error
        # caused by using a binary fraction.
        self._hightime_cache = None

    @singledispatchmethod
    @classmethod
    def _to_offset(cls, value: object) -> TimeDelta:
        raise invalid_arg_type("value", "datetime", value)

    # Python 3.9: pass the type to register() in order to work around
    # https://github.com/python/cpython/issues/86153 - singledispatchmethod raises an error when
    # relying on a forward declaration
    @_to_offset.register(ht.datetime)
    @classmethod
    def _(cls, value: ht.datetime) -> TimeDelta:
        if value.tzinfo != dt.timezone.utc:
            raise ValueError("The tzinfo must be datetime.timezone.utc.")
        return TimeDelta(value - _HT_EPOCH_1904)

    @_to_offset.register(dt.datetime)
    @classmethod
    def _(cls, value: dt.datetime) -> TimeDelta:
        if value.tzinfo != dt.timezone.utc:
            raise ValueError("The tzinfo must be datetime.timezone.utc.")
        return TimeDelta(value - _DT_EPOCH_1904)

    @_to_offset.register(type(None))
    @classmethod
    def _(cls, value: None) -> TimeDelta:
        return TimeDelta()

    @classmethod
    def from_ticks(cls, ticks: SupportsIndex) -> Self:
        """Create an DateTime from a 128-bit fixed point number expressed as an integer."""
        self = cls.__new__(cls)
        self._offset = TimeDelta.from_ticks(ticks)
        self._hightime_cache = None
        return self

    @classmethod
    def from_tuple(cls, value: TimeValueTuple) -> Self:
        """Create a DateTime from whole and fractional seconds as 64-bit ints."""
        self = cls.__new__(cls)
        self._offset = TimeDelta.from_tuple(value)
        self._hightime_cache = None
        return self

    @classmethod
    def from_offset(cls, offset: TimeDelta) -> Self:
        """Create an DateTime from a TimeValue offset from the epoch, Jan 1, 1904."""
        self = cls.__new__(cls)
        self._offset = offset
        self._hightime_cache = None
        return self

    def _to_datetime_datetime(self) -> dt.datetime:
        """Return self as a :any:`datetime.datetime`."""
        return _DT_EPOCH_1904 + self._offset._to_datetime_timedelta()

    def _to_hightime_datetime(self) -> ht.datetime:
        """Return self as a :any:`hightime.datetime`."""
        if self._hightime_cache is None:
            self._hightime_cache = _HT_EPOCH_1904 + self._offset._to_hightime_timedelta()
        return self._hightime_cache

    # Calculating the year/month/day requires knowledge of leap years, days per month, etc., so
    # defer to hightime.datetime.
    @property
    def year(self) -> int:
        """The year."""
        return self._to_hightime_datetime().year

    @property
    def month(self) -> int:
        """The month, between 1 and 12 inclusive."""
        return self._to_hightime_datetime().month

    @property
    def day(self) -> int:
        """The day of the month, between 1 and 31 inclusive."""
        return self._to_hightime_datetime().day

    # The hour/minute/second properties currently assume that tzinfo.utcoffset() == 0, which is
    # true because this class only supports UTC.
    @property
    def hour(self) -> int:
        """The hour, between 0 and 23 inclusive."""
        return self._offset.seconds // 3600

    @property
    def minute(self) -> int:
        """The minute, between 0 and 59 inclusive."""
        return (self._offset.seconds // 60) % 60

    @property
    def second(self) -> int:
        """The second, between 0 and 59 inclusive."""
        return self._offset.seconds % 60

    @property
    def microsecond(self) -> int:
        """The microsecond, between 0 and 999_999 inclusive."""
        return self._offset.microseconds

    @property
    def femtosecond(self) -> int:
        """The femtosecond, between 0 and 999_999_999 inclusive."""
        return self._offset.femtoseconds

    @property
    def yoctosecond(self) -> int:
        """The yoctosecond, between 0 and 999_999_999 inclusive.

        .. warning::
            Because this class uses a 64-bit binary fraction, the smallest time increment it can
            represent is ``1.0 / (1 << 64)`` seconds, which is about 54210 yoctoseconds.
        """
        return self._offset.yoctoseconds

    @property
    def ticks(self) -> int:
        """The number of ticks since the epoch, Jan 1, 1904."""
        return self._offset.ticks

    @property
    def tzinfo(self) -> dt.tzinfo | None:
        """The time zone."""
        return dt.timezone.utc

    def to_tuple(self) -> TimeValueTuple:
        """Convert to the number of whole and fractional seconds since the epoch, Jan 1, 1904."""
        return self._offset.to_tuple()

    @classmethod
    def now(cls, tz: dt.tzinfo | None = None) -> Self:
        """Return the current absolute time."""
        if tz != dt.timezone.utc:
            raise invalid_arg_value("tz", "datetime.timezone.utc", tz)
        return cls(ht.datetime.now(tz))

    def __add__(self, value: TimeDelta | _OtherTimeDelta, /) -> DateTime:
        """Return self+value."""
        if isinstance(value, TimeDelta):
            return self.__class__.from_offset(self._offset + value)
        elif isinstance(value, _OTHER_TIMEDELTA_TUPLE):
            return self + TimeDelta(value)
        else:
            return NotImplemented

    __radd__ = __add__

    @overload
    def __sub__(  # noqa: D105 - missing docstring for magic method
        self, value: DateTime | _OtherDateTime, /
    ) -> TimeDelta: ...
    @overload
    def __sub__(  # noqa: D105 - missing docstring for magic method
        self, value: TimeDelta | _OtherTimeDelta, /
    ) -> DateTime: ...

    def __sub__(
        self, value: DateTime | _OtherDateTime | TimeDelta | _OtherTimeDelta, /
    ) -> TimeDelta | DateTime:
        """Return self-value."""
        if isinstance(value, DateTime):
            return self._offset - value._offset
        elif isinstance(value, _OTHER_DATETIME_TUPLE):
            return self - self.__class__(value)
        elif isinstance(value, TimeDelta):
            return self.__class__.from_offset(self._offset - value)
        elif isinstance(value, _OTHER_TIMEDELTA_TUPLE):
            return self - TimeDelta(value)
        else:
            return NotImplemented

    @overload
    def __rsub__(  # noqa: D105 - missing docstring for magic method
        self, value: DateTime | _OtherDateTime, /
    ) -> TimeDelta: ...
    @overload
    def __rsub__(  # noqa: D105 - missing docstring for magic method
        self, value: TimeDelta | _OtherTimeDelta, /
    ) -> DateTime: ...

    # AB#3153350 - nitypes.bintime.DateTime signatures of __sub__ and __rsub__ are unsafely
    # overlapping
    def __rsub__(  # type: ignore[misc]
        self, value: DateTime | _OtherDateTime | TimeDelta | _OtherTimeDelta, /
    ) -> TimeDelta | DateTime:
        """Return value-self."""
        if isinstance(value, DateTime):
            return value._offset - self._offset
        elif isinstance(value, _OTHER_DATETIME_TUPLE):
            return self.__class__(value) - self
        elif isinstance(value, TimeDelta):
            return self.__class__.from_offset(value - self._offset)
        elif isinstance(value, _OTHER_TIMEDELTA_TUPLE):
            return TimeDelta(value) - self
        else:
            return NotImplemented

    # In comparison operators, always promote to the more precise data type (dt -> bt, bt -> ht).
    def __lt__(self, value: DateTime | _OtherDateTime, /) -> bool:
        """Return self<value."""
        if isinstance(value, self.__class__):
            return self._offset < value._offset
        elif isinstance(value, ht.datetime):
            return self._to_hightime_datetime() < value
        elif isinstance(value, dt.datetime):
            return self < self.__class__(value)
        else:
            return NotImplemented

    def __le__(self, value: DateTime | _OtherDateTime, /) -> bool:
        """Return self<=value."""
        if isinstance(value, self.__class__):
            return self._offset <= value._offset
        elif isinstance(value, ht.datetime):
            return self._to_hightime_datetime() <= value
        elif isinstance(value, dt.datetime):
            return self <= self.__class__(value)
        else:
            return NotImplemented

    def __eq__(self, value: object, /) -> bool:
        """Return self==value."""
        if isinstance(value, self.__class__):
            return self._offset == value._offset
        elif isinstance(value, ht.datetime):
            return self._to_hightime_datetime() == value
        elif isinstance(value, dt.datetime):
            return self == self.__class__(value)
        else:
            return NotImplemented

    def __gt__(self, value: DateTime | _OtherDateTime, /) -> bool:
        """Return self<value."""
        if isinstance(value, self.__class__):
            return self._offset > value._offset
        elif isinstance(value, ht.datetime):
            return self._to_hightime_datetime() > value
        elif isinstance(value, dt.datetime):
            return self > self.__class__(value)
        else:
            return NotImplemented

    def __ge__(self, value: DateTime | _OtherDateTime, /) -> bool:
        """Return self>=value."""
        if isinstance(value, self.__class__):
            return self._offset >= value._offset
        elif isinstance(value, ht.datetime):
            return self._to_hightime_datetime() >= value
        elif isinstance(value, dt.datetime):
            return self >= self.__class__(value)
        else:
            return NotImplemented

    def __hash__(self) -> int:
        """Return hash(self)."""
        return hash(self._offset)

    def __reduce__(self) -> tuple[Any, ...]:
        """Return object state for pickling."""
        return (self.__class__.from_ticks, (self._offset._ticks,))

    def __str__(self) -> str:
        """Return repr(self)."""
        return str(self._to_hightime_datetime())

    def __repr__(self) -> str:
        """Return repr(self)."""
        args: list[int | str] = [self.year, self.month, self.day, self.hour, self.minute]
        # Only display sub-minute fields if they aren't 0.
        if self.yoctosecond:
            args.extend([self.second, self.microsecond, self.femtosecond, self.yoctosecond])
        elif self.femtosecond:
            args.extend([self.second, self.microsecond, self.femtosecond])
        elif self.microsecond:
            args.extend([self.second, self.microsecond])
        elif self.second:
            args.append(self.second)
        args.append("tzinfo=datetime.timezone.utc")
        return f"{self.__class__.__module__}.{self.__class__.__name__}({', '.join(map(str, args))})"


# These have to be within dt.datetime.max/min or else delegating to dt.datetime or ht.datetime for
# year/month/day, str(), repr(), etc. will fail. Use ticks to specify the maximum fractional second
# without rounding up to MAXYEAR+1.
DateTime.max = DateTime(
    dt.MAXYEAR,
    12,
    31,
    23,
    59,
    59,
    tzinfo=dt.timezone.utc,
) + TimeDelta.from_ticks(0xFFFF_FFFF_FFFF_FFFF)
DateTime.min = DateTime(dt.MINYEAR, 1, 1, tzinfo=dt.timezone.utc)



from __future__ import annotations

import operator
from collections.abc import Iterator, Mapping, MutableMapping

from nitypes.waveform.typing import ExtendedPropertyValue

# Extended property keys
CHANNEL_NAME = "NI_ChannelName"
LINE_NAMES = "NI_LineNames"
UNIT_DESCRIPTION = "NI_UnitDescription"


class ExtendedPropertyDictionary(MutableMapping[str, ExtendedPropertyValue]):
    """A dictionary of extended properties.

    .. note::
        Data stored in the extended properties dictionary may not be encrypted when you send it
        over the network or write it to a TDMS file.
    """

    __slots__ = ["_properties"]

    def __init__(self, properties: Mapping[str, ExtendedPropertyValue] | None = None, /) -> None:
        """Initialize a new ExtendedPropertyDictionary."""
        self._properties: dict[str, ExtendedPropertyValue] = {}
        if properties is not None:
            self._properties.update(properties)

    def __len__(self) -> int:
        """Return len(self)."""
        return len(self._properties)

    def __iter__(self) -> Iterator[str]:
        """Implement iter(self)."""
        return iter(self._properties)

    def __contains__(self, value: object, /) -> bool:
        """Implement value in self."""
        return operator.contains(self._properties, value)

    def __getitem__(self, key: str, /) -> ExtendedPropertyValue:
        """Get self[key]."""
        return operator.getitem(self._properties, key)

    def __setitem__(self, key: str, value: ExtendedPropertyValue, /) -> None:
        """Set self[key] to value."""
        operator.setitem(self._properties, key, value)

    def __delitem__(self, key: str, /) -> None:
        """Delete self[key]."""
        operator.delitem(self._properties, key)

    def _merge(self, other: ExtendedPropertyDictionary) -> None:
        for key, value in other.items():
            self._properties.setdefault(key, value)

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"{self.__class__.__module__}.{self.__class__.__name__}({self._properties!r})"



from __future__ import annotations

from collections.abc import Iterable, Sequence
from enum import Enum
from typing import TYPE_CHECKING, final

from nitypes._arguments import validate_unsupported_arg
from nitypes._exceptions import invalid_arg_type
from nitypes.time._types import ANY_DATETIME_TUPLE
from nitypes.waveform._exceptions import create_sample_interval_mode_mismatch_error
from nitypes.waveform._timing._sample_interval._base import SampleIntervalStrategy
from nitypes.waveform._timing._sample_interval._mode import SampleIntervalMode
from nitypes.waveform.errors import TimingMismatchError
from nitypes.waveform.typing import (
    TSampleInterval_co,
    TTimeOffset_co,
    TTimestamp,
    TTimestamp_co,
)

if TYPE_CHECKING:
    from nitypes.waveform._timing._timing import Timing  # circular import


class _Direction(Enum):
    INCREASING = -1
    UNKNOWN = 0
    DECREASING = 1


def _are_timestamps_monotonic(timestamps: Sequence[TTimestamp_co]) -> bool:
    direction = _Direction.UNKNOWN
    for i in range(1, len(timestamps)):
        comparison = _get_direction(timestamps[i - 1], timestamps[i])
        if comparison == _Direction.UNKNOWN:
            continue

        if direction == _Direction.UNKNOWN:
            direction = comparison
        elif comparison != direction:
            return False
    return True


def _get_direction(left: TTimestamp, right: TTimestamp) -> _Direction:
    # Work around https://github.com/python/mypy/issues/18203
    if left < right:  # type: ignore[operator]
        return _Direction.INCREASING
    if right < left:  # type: ignore[operator]
        return _Direction.DECREASING
    return _Direction.UNKNOWN


@final
class IrregularSampleIntervalStrategy(
    SampleIntervalStrategy[TTimestamp_co, TTimeOffset_co, TSampleInterval_co]
):
    """Implements SampleIntervalMode.IRREGULAR specific behavior."""

    def validate_init_args(  # noqa: D102 - Missing docstring in public method - override
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        sample_interval_mode: SampleIntervalMode,
        timestamp: TTimestamp_co | None,
        time_offset: TTimeOffset_co | None,
        sample_interval: TSampleInterval_co | None,
        timestamps: Sequence[TTimestamp_co] | None,
    ) -> None:
        validate_unsupported_arg("timestamp", timestamp)
        validate_unsupported_arg("time offset", time_offset)
        validate_unsupported_arg("sample interval", sample_interval)
        if not isinstance(timestamps, Sequence) or not all(
            isinstance(ts, ANY_DATETIME_TUPLE) for ts in timestamps
        ):
            raise invalid_arg_type("timestamps", "sequence of datetime objects", timestamps)
        if not _are_timestamps_monotonic(timestamps):
            raise ValueError("The timestamps must be in ascending or descending order.")

    def get_timestamps(  # noqa: D102 - Missing docstring in public method - override
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        start_index: int,
        count: int,
    ) -> Iterable[TTimestamp_co]:
        assert timing._timestamps is not None
        if count > len(timing._timestamps):
            raise ValueError("The count must be less than or equal to the number of timestamps.")
        return timing._timestamps[start_index : start_index + count]

    def append_timestamps(  # noqa: D102 - Missing docstring in public method - override
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        timestamps: Sequence[TTimestamp_co] | None,
    ) -> Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co]:
        assert timing._timestamps is not None

        if timestamps is None:
            raise TimingMismatchError(
                "The timestamps argument is required when appending to a waveform with irregular timing."
            )

        datetime_type = type(timing._timestamps[0]) if timing._timestamps else ANY_DATETIME_TUPLE
        if not all(isinstance(ts, datetime_type) for ts in timestamps):
            raise TypeError(
                "The timestamp data type must match the timing information of the current waveform."
            )

        if len(timestamps) == 0:
            return timing
        else:
            if not isinstance(timestamps, list):
                timestamps = list(timestamps)

            return timing.__class__.create_with_irregular_interval(timing._timestamps + timestamps)

    def append_timing(  # noqa: D102 - Missing docstring in public method - override
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        other: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
    ) -> Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co]:
        if other._sample_interval_mode != SampleIntervalMode.IRREGULAR:
            raise create_sample_interval_mode_mismatch_error()

        assert timing._timestamps is not None and other._timestamps is not None

        if len(timing._timestamps) == 0:
            return other
        elif len(other._timestamps) == 0:
            return timing
        else:
            # The constructor will verify that the combined list of timestamps is monotonic. This is
            # not optimal for a large number of appends.
            return timing.__class__.create_with_irregular_interval(
                timing._timestamps + other._timestamps
            )



from __future__ import annotations

from enum import Enum


class SampleIntervalMode(Enum):
    """The sample interval mode that specifies how the waveform is sampled."""

    NONE = 0
    """No sample interval."""

    REGULAR = 1
    """Regular sample interval."""

    IRREGULAR = 2
    """Irregular sample interval."""



"""Digital waveform data types."""

from nitypes.waveform._digital._signal import DigitalWaveformSignal
from nitypes.waveform._digital._signal_collection import DigitalWaveformSignalCollection
from nitypes.waveform._digital._state import DigitalState
from nitypes.waveform._digital._waveform import (
    DigitalWaveform,
    DigitalWaveformFailure,
    DigitalWaveformTestResult,
)

__all__ = [
    "DigitalState",
    "DigitalWaveform",
    "DigitalWaveformFailure",
    "DigitalWaveformSignal",
    "DigitalWaveformSignalCollection",
    "DigitalWaveformTestResult",
]



"""Type aliases and type variables for time types."""

from __future__ import annotations

import datetime as dt
from typing import TypeVar, Union

import hightime as ht
from typing_extensions import TypeAlias

import nitypes.bintime as bt

AnyDateTime: TypeAlias = Union[bt.DateTime, dt.datetime, ht.datetime]
"""Type alias for any ``datetime`` class.

This type alias is a union of the following types:

* :class:`nitypes.bintime.DateTime`
* :class:`datetime.datetime`
* :class:`hightime.datetime`
"""

TDateTime = TypeVar("TDateTime", bound=AnyDateTime)
"""Type variable with a bound of :any:`AnyDateTime`."""

AnyTimeDelta: TypeAlias = Union[bt.TimeDelta, dt.timedelta, ht.timedelta]
"""Type alias for any ``timedelta`` class.

This type alias is a union of the following types:

* :class:`nitypes.bintime.TimeDelta`
* :class:`datetime.timedelta`
* :class:`hightime.timedelta`
"""

TTimeDelta = TypeVar("TTimeDelta", bound=AnyTimeDelta)
"""Type variable with a bound of :any:`AnyTimeDelta`."""



from __future__ import annotations

import operator
from typing import SupportsFloat, SupportsIndex

import numpy as np
import numpy.typing as npt

from nitypes._exceptions import (
    invalid_arg_type,
    invalid_arg_value,
    unsupported_arg,
    unsupported_dtype,
)
from nitypes._numpy import isdtype as _np_isdtype

# Some of these doctests use types introduced in NumPy 2.0 (np.long and np.ulong) or highlight
# formatting differences between NumPy 1.x and 2.x (e.g. dtype=int32, 1.23 vs. np.float64(1.23)).
__doctest_requires__ = {("arg_to_float", "is_dtype", "validate_dtype"): ["numpy>=2.0"]}


def arg_to_float(
    arg_description: str, value: SupportsFloat | None, default_value: float | None = None
) -> float:
    """Convert an argument to a float.

    >>> arg_to_float("xyz", 1.234)
    1.234
    >>> arg_to_float("xyz", 1234)
    1234.0
    >>> arg_to_float("xyz", np.float64(1.234))
    np.float64(1.234)
    >>> arg_to_float("xyz", np.float32(1.234))  # doctest: +ELLIPSIS
    1.233999...
    >>> arg_to_float("xyz", 1.234, 5.0)
    1.234
    >>> arg_to_float("xyz", None, 5.0)
    5.0
    >>> arg_to_float("xyz", None)
    Traceback (most recent call last):
    ...
    TypeError: The xyz must be a floating point number.
    <BLANKLINE>
    Provided value: None
    >>> arg_to_float("xyz", "1.234")
    Traceback (most recent call last):
    ...
    TypeError: The xyz must be a floating point number.
    <BLANKLINE>
    Provided value: '1.234'
    """
    if value is None:
        if default_value is None:
            raise invalid_arg_type(arg_description, "floating point number", value)
        value = default_value

    if not isinstance(value, float):
        try:
            # Use value.__float__() because float(value) also accepts strings.
            return value.__float__()
        except Exception:
            raise invalid_arg_type(arg_description, "floating point number", value) from None

    return value


def arg_to_int(
    arg_description: str, value: SupportsIndex | None, default_value: int | None = None
) -> int:
    """Convert an argument to a signed integer.

    >>> arg_to_int("xyz", 1234)
    1234
    >>> arg_to_int("xyz", 1234, -1)
    1234
    >>> arg_to_int("xyz", None, -1)
    -1
    >>> arg_to_int("xyz", None)
    Traceback (most recent call last):
    ...
    TypeError: The xyz must be an integer.
    <BLANKLINE>
    Provided value: None
    >>> arg_to_int("xyz", 1.234)
    Traceback (most recent call last):
    ...
    TypeError: The xyz must be an integer.
    <BLANKLINE>
    Provided value: 1.234
    >>> arg_to_int("xyz", "1234")
    Traceback (most recent call last):
    ...
    TypeError: The xyz must be an integer.
    <BLANKLINE>
    Provided value: '1234'
    """
    if value is None:
        if default_value is None:
            raise invalid_arg_type(arg_description, "integer", value)
        value = default_value

    if not isinstance(value, int):
        try:
            return operator.index(value)
        except Exception:
            raise invalid_arg_type(arg_description, "integer", value) from None

    return value


def arg_to_uint(
    arg_description: str, value: SupportsIndex | None, default_value: int | None = None
) -> int:
    """Convert an argument to an unsigned integer.

    >>> arg_to_uint("xyz", 1234)
    1234
    >>> arg_to_uint("xyz", 1234, 5000)
    1234
    >>> arg_to_uint("xyz", None, 5000)
    5000
    >>> arg_to_uint("xyz", -1234)
    Traceback (most recent call last):
    ...
    ValueError: The xyz must be a non-negative integer.
    <BLANKLINE>
    Provided value: -1234
    >>> arg_to_uint("xyz", "1234")
    Traceback (most recent call last):
    ...
    TypeError: The xyz must be an integer.
    <BLANKLINE>
    Provided value: '1234'
    """
    value = arg_to_int(arg_description, value, default_value)
    if value < 0:
        raise invalid_arg_value(arg_description, "non-negative integer", value)
    return value


def is_dtype(dtype: npt.DTypeLike, supported_dtypes: tuple[npt.DTypeLike, ...]) -> bool:
    """Check a dtype-like object against a tuple of supported dtype-like objects.

    Unlike :any:`numpy.isdtype`, this function supports structured data types.

    >>> is_dtype(np.float64, (np.float64, np.intc, np.long,))
    True
    >>> is_dtype("float64", (np.float64, np.intc, np.long,))
    True
    >>> is_dtype(np.float64, (np.byte, np.short, np.intc, np.int_, np.long, np.longlong))
    False
    >>> a_type = np.dtype([('a', np.int32)])
    >>> b_type = np.dtype([('b', np.int32)])
    >>> is_dtype(a_type, (np.float64, np.int32, a_type,))
    True
    >>> is_dtype(b_type, (np.float64, np.int32, a_type,))
    False
    >>> is_dtype("i2, i2", (np.float64, np.int32, a_type,))
    False
    >>> is_dtype("i4", (np.float64, np.int32, a_type,))
    True
    """
    if not isinstance(dtype, (type, np.dtype)):
        dtype = np.dtype(dtype)

    if isinstance(dtype, np.dtype) and dtype.fields:
        return dtype in supported_dtypes

    return _np_isdtype(dtype, supported_dtypes)


def validate_dtype(dtype: npt.DTypeLike, supported_dtypes: tuple[npt.DTypeLike, ...]) -> None:
    """Validate a dtype-like object against a tuple of supported dtype-like objects.

    >>> validate_dtype(np.float64, (np.float64, np.intc, np.long,))
    >>> validate_dtype("float64", (np.float64, np.intc, np.long,))
    >>> validate_dtype(np.float64, (np.byte, np.short, np.intc, np.int_, np.long, np.longlong))
    Traceback (most recent call last):
    ...
    TypeError: The requested data type is not supported.
    <BLANKLINE>
    Data type: float64
    Supported data types: int8, int16, int32, int64
    >>> a_type = np.dtype([('a', np.int32)])
    >>> b_type = np.dtype([('b', np.int32)])
    >>> validate_dtype(a_type, (np.float64, np.int32, a_type,))
    >>> validate_dtype(b_type, (np.float64, np.int32, a_type,))
    Traceback (most recent call last):
    ...
    TypeError: The requested data type is not supported.
    <BLANKLINE>
    Data type: [('b', '<i4')]
    Supported data types: float64, int32, void32
    """
    if not is_dtype(dtype, supported_dtypes):
        if not isinstance(dtype, (type, np.dtype)):
            dtype = np.dtype(dtype)
        raise unsupported_dtype("requested data type", dtype, supported_dtypes)


def validate_unsupported_arg(arg_description: str, value: object) -> None:
    """Validate that an unsupported argument is None."""
    if value is not None:
        raise unsupported_arg(arg_description, value)



"""Custom warning classes for waveforms."""

from __future__ import annotations


class ScalingMismatchWarning(RuntimeWarning):
    """Warning used when appending waveforms with mismatched scaling information."""

    pass


class TimingMismatchWarning(RuntimeWarning):
    """Warning used when appending waveforms with mismatched timing information."""

    pass



"""Sample interval strategies for waveform timing."""

from typing import Any

from nitypes._exceptions import invalid_arg_value
from nitypes.waveform._timing._sample_interval._base import SampleIntervalStrategy
from nitypes.waveform._timing._sample_interval._irregular import (
    IrregularSampleIntervalStrategy,
)
from nitypes.waveform._timing._sample_interval._mode import SampleIntervalMode
from nitypes.waveform._timing._sample_interval._none import NoneSampleIntervalStrategy
from nitypes.waveform._timing._sample_interval._regular import (
    RegularSampleIntervalStrategy,
)

__all__ = [
    "create_sample_interval_strategy",
    "IrregularSampleIntervalStrategy",
    "NoneSampleIntervalStrategy",
    "RegularSampleIntervalStrategy",
    "SampleIntervalMode",
    "SampleIntervalStrategy",
]


def create_sample_interval_strategy(
    sample_interval_mode: SampleIntervalMode,
) -> SampleIntervalStrategy[Any, Any, Any]:
    """Create a sample interval strategy for the specified mode."""
    strategy_type = _SAMPLE_INTERVAL_STRATEGY_TYPE_FOR_MODE.get(sample_interval_mode)
    if strategy_type is None:
        raise invalid_arg_value(
            "sample interval mode", "SampleIntervalMode object", sample_interval_mode
        )
    return strategy_type()


_SAMPLE_INTERVAL_STRATEGY_TYPE_FOR_MODE: dict[
    SampleIntervalMode, type[SampleIntervalStrategy[Any, Any, Any]]
] = {
    SampleIntervalMode.NONE: NoneSampleIntervalStrategy,
    SampleIntervalMode.REGULAR: RegularSampleIntervalStrategy,
    SampleIntervalMode.IRREGULAR: IrregularSampleIntervalStrategy,
}



from __future__ import annotations

import warnings
from collections.abc import Generator, Iterable, Sequence
from typing import TYPE_CHECKING, cast, final

from nitypes._arguments import validate_unsupported_arg
from nitypes._exceptions import add_note, invalid_arg_type
from nitypes.time._types import ANY_DATETIME_TUPLE, ANY_TIMEDELTA_TUPLE
from nitypes.waveform._exceptions import (
    create_no_timestamp_information_error,
    create_sample_interval_mode_mismatch_error,
)
from nitypes.waveform._timing._sample_interval._base import SampleIntervalStrategy
from nitypes.waveform._timing._sample_interval._mode import SampleIntervalMode
from nitypes.waveform._warnings import sample_interval_mismatch
from nitypes.waveform.typing import TSampleInterval_co, TTimeOffset_co, TTimestamp_co

if TYPE_CHECKING:
    from nitypes.waveform._timing._timing import Timing  # circular import


@final
class RegularSampleIntervalStrategy(
    SampleIntervalStrategy[TTimestamp_co, TTimeOffset_co, TSampleInterval_co]
):
    """Implements SampleIntervalMode.REGULAR specific behavior."""

    def validate_init_args(  # noqa: D102 - Missing docstring in public method - override
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        sample_interval_mode: SampleIntervalMode,
        timestamp: TTimestamp_co | None,
        time_offset: TTimeOffset_co | None,
        sample_interval: TSampleInterval_co | None,
        timestamps: Sequence[TTimestamp_co] | None,
    ) -> None:
        if not isinstance(timestamp, (ANY_DATETIME_TUPLE, type(None))):
            raise invalid_arg_type("timestamp", "datetime or None", timestamp)
        if not isinstance(time_offset, (ANY_TIMEDELTA_TUPLE, type(None))):
            raise invalid_arg_type("time offset", "timedelta or None", time_offset)
        if not isinstance(sample_interval, ANY_TIMEDELTA_TUPLE):
            raise invalid_arg_type("sample interval", "timedelta", sample_interval)
        validate_unsupported_arg("timestamps", timestamps)

    def get_timestamps(  # noqa: D102 - Missing docstring in public method - override
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        start_index: int,
        count: int,
    ) -> Iterable[TTimestamp_co]:
        if timing.has_timestamp:
            return self._generate_regular_timestamps(timing, start_index, count)
        raise create_no_timestamp_information_error()

    def _generate_regular_timestamps(
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        start_index: int,
        count: int,
    ) -> Generator[TTimestamp_co]:
        sample_interval = timing.sample_interval
        # Work around https://github.com/python/mypy/issues/18203
        timestamp = timing.start_time + start_index * sample_interval  # type: ignore[operator]
        for i in range(count):
            if i != 0:
                timestamp += sample_interval  # type: ignore[operator]
            yield cast(TTimestamp_co, timestamp)

    def append_timestamps(  # noqa: D102 - Missing docstring in public method - override
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        timestamps: Sequence[TTimestamp_co] | None,
    ) -> Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co]:
        try:
            validate_unsupported_arg("timestamps", timestamps)
        except (TypeError, ValueError) as e:
            add_note(e, f"Sample interval mode: {timing.sample_interval_mode}")
            raise
        return timing

    def append_timing(  # noqa: D102 - Missing docstring in public method - override
        self,
        timing: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
        other: Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co],
    ) -> Timing[TTimestamp_co, TTimeOffset_co, TSampleInterval_co]:
        if other._sample_interval_mode not in (SampleIntervalMode.NONE, SampleIntervalMode.REGULAR):
            raise create_sample_interval_mode_mismatch_error()
        if timing._sample_interval != other._sample_interval:
            warnings.warn(sample_interval_mismatch())
        return timing



from __future__ import annotations

from collections.abc import Sequence

import numpy as np
import numpy.typing as npt

from nitypes.waveform.typing import AnyDigitalPort


def bit_mask(n: int, /) -> int:
    """Return the bit mask with the lower n bits set.

    >>> bit_mask(0)
    0
    >>> bit_mask(4)
    15
    >>> bit_mask(9)
    511
    >>> bit_mask(32)
    4294967295
    >>> bit_mask(-1)
    Traceback (most recent call last):
    ...
    ValueError: The number of bits must be a non-negative integer.
    <BLANKLINE>
    Number of bits: -1
    """
    if n < 0:
        raise ValueError(
            "The number of bits must be a non-negative integer.\n\n" f"Number of bits: {n}"
        )
    return (1 << n) - 1


def get_port_dtype(mask: int | Sequence[int]) -> np.dtype[AnyDigitalPort]:
    """Return the NumPy port dtype for the given mask.

    >>> get_port_dtype(0xF)
    dtype('uint8')
    >>> get_port_dtype(0x100)
    dtype('uint16')
    >>> get_port_dtype(0xDEADBEEF)
    dtype('uint32')
    >>> get_port_dtype(0x1_0000_0000)
    Traceback (most recent call last):
    ...
    ValueError: The mask must be an unsigned 8-, 16-, or 32-bit integer.
    <BLANKLINE>
    Mask: 4294967296
    >>> get_port_dtype([0x0F, 0xF0])
    dtype('uint8')
    >>> get_port_dtype([0x100, 0x01])
    dtype('uint16')
    >>> get_port_dtype([0x01, 0x100])
    dtype('uint16')
    >>> get_port_dtype([0xDEADBEEF])
    dtype('uint32')
    """
    if isinstance(mask, Sequence):
        return max((_get_port_dtype(m) for m in mask), key=lambda d: d.itemsize)
    else:
        return _get_port_dtype(mask)


def _get_port_dtype(mask: int) -> np.dtype[AnyDigitalPort]:
    if (mask & 0xFF) == mask:
        return np.dtype(np.uint8)
    elif (mask & 0xFFFF) == mask:
        return np.dtype(np.uint16)
    elif (mask & 0xFFFFFFFF) == mask:
        return np.dtype(np.uint32)
    else:
        raise ValueError(
            "The mask must be an unsigned 8-, 16-, or 32-bit integer.\n\n" f"Mask: {mask}"
        )


def port_to_line_data(port_data: npt.NDArray[AnyDigitalPort], mask: int) -> npt.NDArray[np.uint8]:
    """Convert a 1D array of port data to a 2D array of line data, using the specified mask.

    >>> port_to_line_data(np.array([0,1,2,3], np.uint8), 0xFF)
    array([[0, 0, 0, 0, 0, 0, 0, 0],
           [1, 0, 0, 0, 0, 0, 0, 0],
           [0, 1, 0, 0, 0, 0, 0, 0],
           [1, 1, 0, 0, 0, 0, 0, 0]], dtype=uint8)
    >>> port_to_line_data(np.array([0,1,2,3], np.uint8), 0x3)
    array([[0, 0],
           [1, 0],
           [0, 1],
           [1, 1]], dtype=uint8)
    >>> port_to_line_data(np.array([0,1,2,3], np.uint8), 0x2)
    array([[0],
           [0],
           [1],
           [1]], dtype=uint8)
    >>> port_to_line_data(np.array([0,1,2,3], np.uint8), 0)
    array([], shape=(4, 0), dtype=uint8)
    >>> port_to_line_data(np.array([0x12000000,0xFE000000], np.uint32), 0xFF000000)
    array([[0, 1, 0, 0, 1, 0, 0, 0],
           [0, 1, 1, 1, 1, 1, 1, 1]], dtype=uint8)
    """
    port_size = port_data.dtype.itemsize * 8
    line_data = np.unpackbits(port_data.view(np.uint8), bitorder="little")
    line_data = line_data.reshape(len(port_data), port_size)

    if mask == bit_mask(port_size):
        return line_data
    else:
        return line_data[:, _mask_to_line_indices(mask)]


def _mask_to_line_indices(mask: int, /) -> list[int]:
    """Return the line indices for the given mask.

    >>> _mask_to_line_indices(0xF)
    [0, 1, 2, 3]
    >>> _mask_to_line_indices(0x100)
    [8]
    >>> _mask_to_line_indices(0xDEADBEEF)
    [0, 1, 2, 3, 5, 6, 7, 9, 10, 11, 12, 13, 15, 16, 18, 19, 21, 23, 25, 26, 27, 28, 30, 31]
    >>> _mask_to_line_indices(-1)
    Traceback (most recent call last):
    ...
    ValueError: The mask must be a non-negative integer.
    <BLANKLINE>
    Mask: -1
    """
    if mask < 0:
        raise ValueError("The mask must be a non-negative integer.\n\n" f"Mask: {mask}")
    line_indices = []
    line_index = 0
    while mask != 0:
        if mask & 1:
            line_indices.append(line_index)
        line_index += 1
        mask >>= 1
    return line_indices



from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Any, SupportsIndex, Union, overload

import numpy as np
import numpy.typing as npt
from typing_extensions import TYPE_CHECKING, TypeVar, final, override

from nitypes._numpy import long as _np_long, ulong as _np_ulong
from nitypes.time.typing import AnyDateTime, AnyTimeDelta
from nitypes.waveform._numeric import _TOtherScaled
from nitypes.waveform.typing import ExtendedPropertyValue

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.waveform import NumericWaveform, ScaleMode, Timing
else:
    from nitypes.waveform._numeric import NumericWaveform
    from nitypes.waveform._scaling import ScaleMode
    from nitypes.waveform._timing import Timing


# _TRaw specifies the type of the raw_data array. AnalogWaveform accepts a narrower set of types
# than NumericWaveform.
_TRaw = TypeVar("_TRaw", bound=Union[np.floating, np.integer])
_TOtherRaw = TypeVar("_TOtherRaw", bound=Union[np.floating, np.integer])

# Use the C types here because np.isdtype() considers some of them to be distinct types, even when
# they have the same size (e.g. np.intc vs. np.int_).
_RAW_DTYPES = (
    # Floating point
    np.single,
    np.double,
    # Signed integers
    np.byte,
    np.short,
    np.intc,
    np.int_,
    _np_long,
    np.longlong,
    # Unsigned integers
    np.ubyte,
    np.ushort,
    np.uintc,
    np.uint,
    _np_ulong,
    np.ulonglong,
)

_SCALED_DTYPES = (
    # Floating point
    np.single,
    np.double,
)


@final
class AnalogWaveform(NumericWaveform[_TRaw, np.float64]):
    """An analog waveform, which encapsulates analog data and timing information.

    Constructing
    ^^^^^^^^^^^^

    To construct an analog waveform, use the :class:`AnalogWaveform` class:

    >>> AnalogWaveform()
    nitypes.waveform.AnalogWaveform(0)
    >>> AnalogWaveform(5)
    nitypes.waveform.AnalogWaveform(5, raw_data=array([0., 0., 0., 0., 0.]))

    To construct an analog waveform from a NumPy array, use the :any:`AnalogWaveform.from_array_1d`
    method.

    >>> import numpy as np
    >>> AnalogWaveform.from_array_1d(np.array([1.0, 2.0, 3.0]))
    nitypes.waveform.AnalogWaveform(3, raw_data=array([1., 2., 3.]))

    You can also use :any:`AnalogWaveform.from_array_1d` to construct an analog waveform from a
    sequence, such as a list. In this case, you must specify the NumPy data type.

    >>> AnalogWaveform.from_array_1d([1.0, 2.0, 3.0], np.float64)
    nitypes.waveform.AnalogWaveform(3, raw_data=array([1., 2., 3.]))

    The 2D version, :any:`AnalogWaveform.from_array_2d`, returns multiple waveforms, one for each row of
    data in the array or nested sequence.

    >>> nested_list = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
    >>> AnalogWaveform.from_array_2d(nested_list, np.float64)  # doctest: +NORMALIZE_WHITESPACE
    [nitypes.waveform.AnalogWaveform(3, raw_data=array([1., 2., 3.])),
    nitypes.waveform.AnalogWaveform(3, raw_data=array([4., 5., 6.]))]

    Timing information
    ^^^^^^^^^^^^^^^^^^

    Analog waveforms include timing information, such as the start time and sample interval, to support
    analyzing and visualizing the data.

    You can specify timing information by constructing a :class:`Timing` object and passing it to the
    waveform constructor or factory method:

    >>> import datetime as dt
    >>> wfm = AnalogWaveform(timing=Timing.create_with_regular_interval(
    ...     dt.timedelta(seconds=1e-3), dt.datetime(2024, 12, 31, 23, 59, 59, tzinfo=dt.timezone.utc)
    ... ))
    >>> wfm.timing  # doctest: +NORMALIZE_WHITESPACE
    nitypes.waveform.Timing(nitypes.waveform.SampleIntervalMode.REGULAR,
        timestamp=datetime.datetime(2024, 12, 31, 23, 59, 59, tzinfo=datetime.timezone.utc),
        sample_interval=datetime.timedelta(microseconds=1000))

    You can query the waveform's timing information using the :class:`Timing` object's properties:

    >>> wfm.timing.start_time
    datetime.datetime(2024, 12, 31, 23, 59, 59, tzinfo=datetime.timezone.utc)
    >>> wfm.timing.sample_interval
    datetime.timedelta(microseconds=1000)

    Timing objects are immutable, so you cannot directly set their properties:

    >>> wfm.timing.sample_interval = dt.timedelta(seconds=10e-3)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    AttributeError: ...

    Instead, if you want to modify the timing information for an existing waveform, you can create a new
    timing object and set the :any:`NumericWaveform.timing` property:

    >>> wfm.timing = Timing.create_with_regular_interval(
    ...     dt.timedelta(seconds=1e-3), dt.datetime(2025, 1, 1, tzinfo=dt.timezone.utc)
    ... )
    >>> wfm.timing  # doctest: +NORMALIZE_WHITESPACE
    nitypes.waveform.Timing(nitypes.waveform.SampleIntervalMode.REGULAR,
        timestamp=datetime.datetime(2025, 1, 1, 0, 0, tzinfo=datetime.timezone.utc),
        sample_interval=datetime.timedelta(microseconds=1000))

    Timing objects support time types from the :class:`DateTime`, :any:`hightime`, and
    :any:`nitypes.bintime` modules. If you need the timing information in a specific representation, use
    the conversion methods:

    >>> wfm.timing.to_datetime()  # doctest: +NORMALIZE_WHITESPACE
    nitypes.waveform.Timing(nitypes.waveform.SampleIntervalMode.REGULAR,
        timestamp=datetime.datetime(2025, 1, 1, 0, 0, tzinfo=datetime.timezone.utc),
        sample_interval=datetime.timedelta(microseconds=1000))
    >>> wfm.timing.to_hightime()  # doctest: +NORMALIZE_WHITESPACE
    nitypes.waveform.Timing(nitypes.waveform.SampleIntervalMode.REGULAR,
        timestamp=hightime.datetime(2025, 1, 1, 0, 0, tzinfo=datetime.timezone.utc),
        sample_interval=hightime.timedelta(microseconds=1000))
    >>> wfm.timing.to_bintime()  # doctest: +NORMALIZE_WHITESPACE
    nitypes.waveform.Timing(nitypes.waveform.SampleIntervalMode.REGULAR,
        timestamp=nitypes.bintime.DateTime(2025, 1, 1, 0, 0, tzinfo=datetime.timezone.utc),
        sample_interval=nitypes.bintime.TimeDelta(Decimal('0.000999999999999999966606573')))

    If :any:`NumericWaveform.timing` is not specified for a given waveform, it defaults to the
    :any:`Timing.empty` singleton object.

    >>> AnalogWaveform().timing
    nitypes.waveform.Timing(nitypes.waveform.SampleIntervalMode.NONE)
    >>> AnalogWaveform().timing is Timing.empty
    True

    Accessing unspecified properties of the timing object raises an exception:

    >>> Timing.empty.sample_interval
    Traceback (most recent call last):
    ...
    RuntimeError: The waveform timing does not have a sample interval.

    You can use :any:`Timing.sample_interval_mode` and ``has_*`` properties such as
    :any:`Timing.has_timestamp` to query which properties of the timing object were specified:

    >>> wfm.timing.sample_interval_mode
    <SampleIntervalMode.REGULAR: 1>
    >>> (wfm.timing.has_timestamp, wfm.timing.has_sample_interval)
    (True, True)
    >>> Timing.empty.sample_interval_mode
    <SampleIntervalMode.NONE: 0>
    >>> (Timing.empty.has_timestamp, Timing.empty.has_sample_interval)
    (False, False)

    Scaling analog data
    ^^^^^^^^^^^^^^^^^^^

    By default, analog waveforms contain floating point data in :any:`numpy.float64` format, but they
    can also be used to scale raw integer data to floating-point:

    >>> import numpy as np
    >>> scale_mode = LinearScaleMode(gain=2.0, offset=0.5)
    >>> wfm = AnalogWaveform.from_array_1d([1, 2, 3], np.int32, scale_mode=scale_mode)
    >>> wfm  # doctest: +NORMALIZE_WHITESPACE
    nitypes.waveform.AnalogWaveform(3, int32, raw_data=array([1, 2, 3], dtype=int32),
        scale_mode=nitypes.waveform.LinearScaleMode(2.0, 0.5))
    >>> wfm.raw_data
    array([1, 2, 3], dtype=int32)
    >>> wfm.scaled_data
    array([2.5, 4.5, 6.5])

    Class members
    ^^^^^^^^^^^^^
    """  # noqa: W505 - doc line too long

    @override
    @staticmethod
    def _get_default_raw_dtype() -> type[np.generic] | np.dtype[np.generic]:
        return np.float64

    @override
    @staticmethod
    def _get_default_scaled_dtype() -> type[np.generic] | np.dtype[np.generic]:
        return np.float64

    @override
    @staticmethod
    def _get_supported_raw_dtypes() -> tuple[npt.DTypeLike, ...]:
        return _RAW_DTYPES

    @override
    @staticmethod
    def _get_supported_scaled_dtypes() -> tuple[npt.DTypeLike, ...]:
        return _SCALED_DTYPES

    # Override from_array_1d, from_array_2d, and __init__ in order to use overloads to control type
    # inference.
    @overload
    @classmethod
    def from_array_1d(
        cls,
        array: npt.NDArray[_TOtherRaw],
        dtype: None = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> AnalogWaveform[_TOtherRaw]: ...

    @overload
    @classmethod
    def from_array_1d(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: type[_TOtherRaw] | np.dtype[_TOtherRaw],
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> AnalogWaveform[_TOtherRaw]: ...

    @overload
    @classmethod
    def from_array_1d(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: npt.DTypeLike = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> AnalogWaveform[Any]: ...

    @override
    @classmethod
    def from_array_1d(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: npt.DTypeLike = None,
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = None,
        scale_mode: ScaleMode | None = None,
    ) -> AnalogWaveform[Any]:
        """Construct an analog waveform from a one-dimensional array or sequence.

        Args:
            array: The waveform data as a one-dimensional array or a sequence.
            dtype: The NumPy data type for the waveform data. This argument is required
                when array is a sequence.
            copy: Specifies whether to copy the array or save a reference to it.
            start_index: The sample index at which the waveform data begins.
            sample_count: The number of samples in the waveform.
            extended_properties: The extended properties of the waveform.
            timing: The timing information of the waveform.
            scale_mode: The scale mode of the waveform.

        Returns:
            An analog waveform containing the specified data.
        """
        return super().from_array_1d(
            array,
            dtype,
            copy=copy,
            start_index=start_index,
            sample_count=sample_count,
            extended_properties=extended_properties,
            timing=timing,
            scale_mode=scale_mode,
        )

    @overload
    @classmethod
    def from_array_2d(
        cls,
        array: npt.NDArray[_TOtherRaw],
        dtype: None = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> Sequence[AnalogWaveform[_TOtherRaw]]: ...

    @overload
    @classmethod
    def from_array_2d(
        cls,
        array: npt.NDArray[Any] | Sequence[Sequence[Any]],
        dtype: type[_TOtherRaw] | np.dtype[_TOtherRaw],
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> Sequence[AnalogWaveform[_TOtherRaw]]: ...

    @overload
    @classmethod
    def from_array_2d(
        cls,
        array: npt.NDArray[Any] | Sequence[Sequence[Any]],
        dtype: npt.DTypeLike = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> Sequence[AnalogWaveform[Any]]: ...

    @override
    @classmethod
    def from_array_2d(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls,
        array: npt.NDArray[Any] | Sequence[Sequence[Any]],
        dtype: npt.DTypeLike = None,
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = None,
        scale_mode: ScaleMode | None = None,
    ) -> Sequence[AnalogWaveform[Any]]:
        """Construct multiple analog waveforms from a two-dimensional array or nested sequence.

        Args:
            array: The waveform data as a two-dimensional array or a nested sequence.
            dtype: The NumPy data type for the waveform data. This argument is required
                when array is a sequence.
            copy: Specifies whether to copy the array or save a reference to it.
            start_index: The sample index at which the waveform data begins.
            sample_count: The number of samples in the waveform.
            extended_properties: The extended properties of the waveform.
            timing: The timing information of the waveform.
            scale_mode: The scale mode of the waveform.

        Returns:
            A sequence containing an analog waveform for each row of the specified data.

        When constructing multiple waveforms, the same extended properties, timing
        information, and scale mode are applied to all waveforms. Consider assigning
        these properties after construction.
        """
        return super().from_array_2d(
            array,
            dtype,
            copy=copy,
            start_index=start_index,
            sample_count=sample_count,
            extended_properties=extended_properties,
            timing=timing,
            scale_mode=scale_mode,
        )

    __slots__ = ()

    # If neither dtype nor raw_data is specified, _TRaw defaults to np.float64.
    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: AnalogWaveform[np.float64],
        sample_count: SupportsIndex | None = ...,
        dtype: None = ...,
        *,
        raw_data: None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: AnalogWaveform[_TOtherRaw],
        sample_count: SupportsIndex | None = ...,
        dtype: type[_TOtherRaw] | np.dtype[_TOtherRaw] = ...,
        *,
        raw_data: None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: AnalogWaveform[_TOtherRaw],
        sample_count: SupportsIndex | None = ...,
        dtype: None = ...,
        *,
        raw_data: npt.NDArray[_TOtherRaw] = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: AnalogWaveform[Any],
        sample_count: SupportsIndex | None = ...,
        dtype: npt.DTypeLike = ...,
        *,
        raw_data: npt.NDArray[Any] | None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> None: ...

    def __init__(
        self,
        sample_count: SupportsIndex | None = None,
        dtype: npt.DTypeLike = None,
        *,
        raw_data: npt.NDArray[Any] | None = None,
        start_index: SupportsIndex | None = None,
        capacity: SupportsIndex | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        copy_extended_properties: bool = True,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = None,
        scale_mode: ScaleMode | None = None,
    ) -> None:
        """Initialize a new analog waveform.

        Args:
            sample_count: The number of samples in the analog waveform.
            dtype: The NumPy data type for the analog waveform data. If not specified, the data
                type defaults to np.float64.
            raw_data: A NumPy ndarray to use for sample storage. The analog waveform takes ownership
                of this array. If not specified, an ndarray is created based on the specified dtype,
                start index, sample count, and capacity.
            start_index: The sample index at which the analog waveform data begins.
            sample_count: The number of samples in the analog waveform.
            capacity: The number of samples to allocate. Pre-allocating a larger buffer optimizes
                appending samples to the waveform.
            extended_properties: The extended properties of the analog waveform.
            copy_extended_properties: Specifies whether to copy the extended properties or take
                ownership.
            timing: The timing information of the analog waveform.
            scale_mode: The scale mode of the analog waveform.

        Returns:
            An analog waveform.
        """
        return super().__init__(
            sample_count,
            dtype,
            raw_data=raw_data,
            start_index=start_index,
            capacity=capacity,
            extended_properties=extended_properties,
            copy_extended_properties=copy_extended_properties,
            timing=timing,
            scale_mode=scale_mode,
        )

    @override
    def _convert_data(
        self,
        dtype: npt.DTypeLike | type[_TOtherScaled] | np.dtype[_TOtherScaled],
        raw_data: npt.NDArray[_TRaw],
    ) -> npt.NDArray[_TOtherScaled]:
        return raw_data.astype(dtype)



from __future__ import annotations

import numpy as np
from typing_extensions import TypeAlias

CVIAbsoluteTimeBase: TypeAlias = np.void
"""Type alias for the base type of :any:`CVIAbsoluteTimeDType`, which is :any:`numpy.void`."""

CVIAbsoluteTimeDType = np.dtype((CVIAbsoluteTimeBase, [("lsb", np.uint64), ("msb", np.int64)]))
"""NumPy structured data type for a ``CVIAbsoluteTime`` C struct."""

CVITimeIntervalBase: TypeAlias = np.void
"""Type alias for the base type of :any:`CVITimeIntervalDType`, which is :any:`numpy.void`."""

CVITimeIntervalDType = np.dtype((CVITimeIntervalBase, [("lsb", np.uint64), ("msb", np.int64)]))
"""NumPy structured data type for a ``CVITimeInterval`` C struct."""



from __future__ import annotations

import numpy as np

from nitypes._numpy import bool as _np_bool

DIGITAL_PORT_DTYPES = (np.uint8, np.uint16, np.uint32)
"""Tuple of types corresponding to :any:`AnyDigitalPort`."""

DIGITAL_STATE_DTYPES = (_np_bool, np.int8, np.uint8)
"""Tuple of types corresponding to :any:`AnyDigitalState`."""



from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Any, SupportsIndex, Union, overload

import numpy as np
import numpy.typing as npt
from typing_extensions import TYPE_CHECKING, TypeVar, final, override

from nitypes.complex import ComplexInt32Base, ComplexInt32DType, convert_complex
from nitypes.time.typing import AnyDateTime, AnyTimeDelta
from nitypes.waveform._numeric import _TOtherScaled
from nitypes.waveform.typing import ExtendedPropertyValue

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.waveform import NumericWaveform, ScaleMode, Timing
else:
    from nitypes.waveform._numeric import NumericWaveform
    from nitypes.waveform._scaling import ScaleMode
    from nitypes.waveform._timing import Timing

# _TRaw specifies the type of the raw_data array. ComplexWaveform accepts a narrower set of types
# than NumericWaveform. Note that ComplexInt32Base is an alias for np.void, but other structured
# data types are rejected at run time.
_TRaw = TypeVar("_TRaw", bound=Union[np.complexfloating, ComplexInt32Base])
_TOtherRaw = TypeVar("_TOtherRaw", bound=Union[np.complexfloating, ComplexInt32Base])

_RAW_DTYPES = (
    # Complex floating point
    np.csingle,
    np.cdouble,
    # Complex integers
    ComplexInt32DType,
)

_SCALED_DTYPES = (
    # Complex floating point
    np.csingle,
    np.cdouble,
)


@final
class ComplexWaveform(NumericWaveform[_TRaw, np.complex128]):
    """A complex waveform, which encapsulates complex data and timing information.

    Constructing
    ^^^^^^^^^^^^

    To construct a complex waveform, use the :class:`ComplexWaveform` class:

    >>> ComplexWaveform()
    nitypes.waveform.ComplexWaveform(0)
    >>> ComplexWaveform(5)
    nitypes.waveform.ComplexWaveform(5, raw_data=array([0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j]))

    To construct a complex waveform from a NumPy array, use the :any:`ComplexWaveform.from_array_1d`
    method.

    >>> import numpy as np
    >>> ComplexWaveform.from_array_1d(np.array([1+2j, 3+4j, 5+6j]))
    nitypes.waveform.ComplexWaveform(3, raw_data=array([1.+2.j, 3.+4.j, 5.+6.j]))

    You can also use :any:`ComplexWaveform.from_array_1d` to construct a complex waveform from a
    sequence, such as a list. In this case, you must specify the NumPy data type.

    >>> ComplexWaveform.from_array_1d([1+2j, 3+4j, 5+6j], np.complex128)
    nitypes.waveform.ComplexWaveform(3, raw_data=array([1.+2.j, 3.+4.j, 5.+6.j]))

    The 2D version, :any:`ComplexWaveform.from_array_2d`, returns multiple waveforms, one for each row of
    data in the array or nested sequence.

    >>> nested_list = [[1+2j, 3+4j, 5+6j], [7+8j, 9+10j, 11+12j]]
    >>> ComplexWaveform.from_array_2d(nested_list, np.complex128)  # doctest: +NORMALIZE_WHITESPACE
    [nitypes.waveform.ComplexWaveform(3, raw_data=array([1.+2.j, 3.+4.j, 5.+6.j])),
    nitypes.waveform.ComplexWaveform(3, raw_data=array([ 7. +8.j,  9.+10.j, 11.+12.j]))]

    Scaling complex-number data
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^

    Complex waveforms support scaling raw integer data to floating-point. Python and NumPy do not have
    native support for complex integers, so this uses the :any:`ComplexInt32DType` structured data type.

    >>> from nitypes.complex import ComplexInt32DType
    >>> scale_mode = LinearScaleMode(gain=2.0, offset=0.5)
    >>> wfm = ComplexWaveform.from_array_1d([(1, 2), (3, 4)], ComplexInt32DType, scale_mode=scale_mode)
    >>> wfm  # doctest: +NORMALIZE_WHITESPACE
    nitypes.waveform.ComplexWaveform(2, void32, raw_data=array([(1, 2), (3, 4)],
        dtype=[('real', '<i2'), ('imag', '<i2')]),
        scale_mode=nitypes.waveform.LinearScaleMode(2.0, 0.5))
    >>> wfm.raw_data
    array([(1, 2), (3, 4)], dtype=[('real', '<i2'), ('imag', '<i2')])
    >>> wfm.scaled_data
    array([2.5+4.j, 6.5+8.j])

    Timing information
    ^^^^^^^^^^^^^^^^^^

    Complex waveforms have the same timing information as analog waveforms. For more details, see
    :class:`AnalogWaveform`.

    Class members
    ^^^^^^^^^^^^^
    """  # noqa: W505 - doc line too long

    @override
    @staticmethod
    def _get_default_raw_dtype() -> type[np.generic] | np.dtype[np.generic]:
        return np.complex128

    @override
    @staticmethod
    def _get_default_scaled_dtype() -> type[np.generic] | np.dtype[np.generic]:
        return np.complex128

    @override
    @staticmethod
    def _get_supported_raw_dtypes() -> tuple[npt.DTypeLike, ...]:
        return _RAW_DTYPES

    @override
    @staticmethod
    def _get_supported_scaled_dtypes() -> tuple[npt.DTypeLike, ...]:
        return _SCALED_DTYPES

    @overload
    @classmethod
    def from_array_1d(
        cls,
        array: npt.NDArray[_TOtherRaw],
        dtype: None = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> ComplexWaveform[_TOtherRaw]: ...

    @overload
    @classmethod
    def from_array_1d(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: type[_TOtherRaw] | np.dtype[_TOtherRaw],
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> ComplexWaveform[_TOtherRaw]: ...

    @overload
    @classmethod
    def from_array_1d(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: npt.DTypeLike = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> ComplexWaveform[Any]: ...

    @override
    @classmethod
    def from_array_1d(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: npt.DTypeLike = None,
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = None,
        scale_mode: ScaleMode | None = None,
    ) -> ComplexWaveform[Any]:
        """Construct a complex waveform from a one-dimensional array or sequence.

        Args:
            array: The waveform data as a one-dimensional array or a sequence.
            dtype: The NumPy data type for the waveform data. This argument is required
                when array is a sequence.
            copy: Specifies whether to copy the array or save a reference to it.
            start_index: The sample index at which the waveform data begins.
            sample_count: The number of samples in the waveform.
            extended_properties: The extended properties of the waveform.
            timing: The timing information of the waveform.
            scale_mode: The scale mode of the waveform.

        Returns:
            A complex waveform containing the specified data.
        """
        return super().from_array_1d(
            array,
            dtype,
            copy=copy,
            start_index=start_index,
            sample_count=sample_count,
            extended_properties=extended_properties,
            timing=timing,
            scale_mode=scale_mode,
        )

    @overload
    @classmethod
    def from_array_2d(
        cls,
        array: npt.NDArray[_TOtherRaw],
        dtype: None = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> Sequence[ComplexWaveform[_TOtherRaw]]: ...

    @overload
    @classmethod
    def from_array_2d(
        cls,
        array: npt.NDArray[Any] | Sequence[Sequence[Any]],
        dtype: type[_TOtherRaw] | np.dtype[_TOtherRaw],
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> Sequence[ComplexWaveform[_TOtherRaw]]: ...

    @overload
    @classmethod
    def from_array_2d(
        cls,
        array: npt.NDArray[Any] | Sequence[Sequence[Any]],
        dtype: npt.DTypeLike = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> Sequence[ComplexWaveform[Any]]: ...

    @override
    @classmethod
    def from_array_2d(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls,
        array: npt.NDArray[Any] | Sequence[Sequence[Any]],
        dtype: npt.DTypeLike = None,
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = None,
        scale_mode: ScaleMode | None = None,
    ) -> Sequence[ComplexWaveform[Any]]:
        """Construct multiple complex waveforms from a two-dimensional array or nested sequence.

        Args:
            array: The waveform data as a two-dimensional array or a nested sequence.
            dtype: The NumPy data type for the waveform data. This argument is required
                when array is a sequence.
            copy: Specifies whether to copy the array or save a reference to it.
            start_index: The sample index at which the waveform data begins.
            sample_count: The number of samples in the waveform.
            extended_properties: The extended properties of the waveform.
            timing: The timing information of the waveform.
            scale_mode: The scale mode of the waveform.

        Returns:
            A sequence containing a complex waveform for each row of the specified data.

        When constructing multiple waveforms, the same extended properties, timing
        information, and scale mode are applied to all waveforms. Consider assigning
        these properties after construction.
        """
        return super().from_array_2d(
            array,
            dtype,
            copy=copy,
            start_index=start_index,
            sample_count=sample_count,
            extended_properties=extended_properties,
            timing=timing,
            scale_mode=scale_mode,
        )

    __slots__ = ()

    # If neither dtype nor raw_data is specified, _TRaw defaults to np.complex128.
    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: ComplexWaveform[np.complex128],
        sample_count: SupportsIndex | None = ...,
        dtype: None = ...,
        *,
        raw_data: None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: ComplexWaveform[_TOtherRaw],
        sample_count: SupportsIndex | None = ...,
        dtype: type[_TOtherRaw] | np.dtype[_TOtherRaw] = ...,
        *,
        raw_data: None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: ComplexWaveform[_TOtherRaw],
        sample_count: SupportsIndex | None = ...,
        dtype: None = ...,
        *,
        raw_data: npt.NDArray[_TOtherRaw] = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: ComplexWaveform[Any],
        sample_count: SupportsIndex | None = ...,
        dtype: npt.DTypeLike = ...,
        *,
        raw_data: npt.NDArray[Any] | None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
        scale_mode: ScaleMode | None = ...,
    ) -> None: ...

    def __init__(
        self,
        sample_count: SupportsIndex | None = None,
        dtype: npt.DTypeLike = None,
        *,
        raw_data: npt.NDArray[Any] | None = None,
        start_index: SupportsIndex | None = None,
        capacity: SupportsIndex | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        copy_extended_properties: bool = True,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = None,
        scale_mode: ScaleMode | None = None,
    ) -> None:
        """Initialize a new complex waveform.

        Args:
            sample_count: The number of samples in the waveform.
            dtype: The NumPy data type for the waveform data. If not specified, the data
                type defaults to np.complex128.
            raw_data: A NumPy ndarray to use for sample storage. The waveform takes ownership
                of this array. If not specified, an ndarray is created based on the specified dtype,
                start index, sample count, and capacity.
            start_index: The sample index at which the waveform data begins.
            sample_count: The number of samples in the waveform.
            capacity: The number of samples to allocate. Pre-allocating a larger buffer optimizes
                appending samples to the waveform.
            extended_properties: The extended properties of the waveform.
            copy_extended_properties: Specifies whether to copy the extended properties or take
                ownership.
            timing: The timing information of the waveform.
            scale_mode: The scale mode of the waveform.

        Returns:
            A complex waveform.
        """
        return super().__init__(
            sample_count,
            dtype,
            raw_data=raw_data,
            start_index=start_index,
            capacity=capacity,
            extended_properties=extended_properties,
            copy_extended_properties=copy_extended_properties,
            timing=timing,
            scale_mode=scale_mode,
        )

    @override
    def _convert_data(
        self,
        dtype: npt.DTypeLike | type[_TOtherScaled] | np.dtype[_TOtherScaled],
        raw_data: npt.NDArray[_TRaw],
    ) -> npt.NDArray[_TOtherScaled]:
        return convert_complex(dtype, raw_data)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeVar

import numpy as np
import numpy.typing as npt


_ScalarType = TypeVar("_ScalarType", bound=np.generic)


class ScaleMode(ABC):
    """An object that specifies how the waveform is scaled."""

    __slots__ = ()

    @abstractmethod
    def _transform_data(self, data: npt.NDArray[_ScalarType]) -> npt.NDArray[_ScalarType]:
        raise NotImplementedError
